"""Contains tests for the connection classes."""

from asyncio import StreamReader, StreamWriter
from importlib import reload
import logging
import sys
from typing import Final
from unittest.mock import AsyncMock, patch

import pytest
from serial import EIGHTBITS, PARITY_NONE, STOPBITS_ONE, SerialException

import pyplumio.connection
from pyplumio.exceptions import ConnectionFailedError
from pyplumio.protocol import Protocol

HOST: Final = "localhost"
PORT: Final = 8899
DEVICE: Final = "/dev/ttyUSB0"


@pytest.fixture(name="use_fast", params=(True, False))
def fixture_use_fast(request, monkeypatch, caplog):
    """Try with and without serial-asyncio-fast package."""
    if not request.param:
        monkeypatch.setitem(sys.modules, "serial_asyncio_fast", None)

    with caplog.at_level(logging.INFO):
        reload(pyplumio.connection)

    message = "Using pyserial-asyncio-fast in place of pyserial-asyncio"
    assert (message in caplog.text) if request.param else (message not in caplog.text)
    return request.param


@pytest.fixture(name="stream_writer")
def fixture_stream_writer():
    """Return a mock of asyncio stream writer."""
    with patch("asyncio.StreamWriter", autospec=True) as mock_stream_writer:
        yield mock_stream_writer


@pytest.fixture(name="stream_reader")
def fixture_stream_reader():
    """Return a mock of asyncio stream reader."""
    with patch("asyncio.StreamReader", autospec=True) as mock_stream_reader:
        yield mock_stream_reader


@pytest.fixture(name="asyncio_open_connection")
def fixture_asyncio_open_connection(
    stream_reader: StreamReader, stream_writer: StreamWriter
):
    """Bypass opening asyncio connection."""
    with patch(
        "asyncio.open_connection", return_value=(stream_reader, stream_writer)
    ) as mock_connection:
        yield mock_connection


@pytest.fixture(name="serial_asyncio_open_serial_connection")
def fixture_serial_asyncio_open_serial_connection(
    use_fast, stream_reader: StreamReader, stream_writer: StreamWriter
):
    """Bypass opening serial_asyncio connection."""
    module = "serial_asyncio_fast" if use_fast else "serial_asyncio"
    with patch(
        f"{module}.open_serial_connection",
        return_value=(stream_reader, stream_writer),
    ) as mock_connection:
        yield mock_connection


@pytest.fixture(name="mock_protocol")
def fixture_mock_protocol():
    """Return a mock protocol object."""
    return AsyncMock(spec=Protocol)


@pytest.fixture(name="tcp_connection")
def fixture_tcp_connection(mock_protocol) -> pyplumio.connection.TcpConnection:
    """Return at TCP connection object."""
    return pyplumio.connection.TcpConnection(
        host=HOST, port=PORT, timeout=10, protocol=mock_protocol
    )


@pytest.fixture(name="serial_connection")
def fixture_serial_connection(mock_protocol) -> pyplumio.connection.SerialConnection:
    """Return a serial connection object."""
    return pyplumio.connection.SerialConnection(
        device="/dev/ttyUSB0", timeout=10, protocol=mock_protocol
    )


async def test_tcp_connect(mock_protocol, asyncio_open_connection) -> None:
    """Test TCP connection logic."""
    with patch("pyplumio.connection.Connection._reconnect") as mock_reconnect:
        tcp_connection = pyplumio.connection.TcpConnection(
            host=HOST,
            port=PORT,
            timeout=10,
            protocol=mock_protocol,
            reconnect_on_failure=True,
        )

        assert tcp_connection.protocol == mock_protocol
        mock_protocol.on_connection_lost.add.assert_called_once_with(mock_reconnect)

    await tcp_connection.connect()
    asyncio_open_connection.assert_called_once_with(host=HOST, port=PORT, timeout=10)
    await tcp_connection.close()

    # Raise custom exception on connection failure.
    tcp_connection = pyplumio.connection.TcpConnection(
        host=HOST,
        port=PORT,
        timeout=10,
        protocol=mock_protocol,
        reconnect_on_failure=False,
    )

    asyncio_open_connection.side_effect = OSError
    with pytest.raises(ConnectionFailedError):
        await tcp_connection.connect()


async def test_serial_connect(
    mock_protocol, serial_asyncio_open_serial_connection
) -> None:
    """Test a serial connection logic."""
    serial_connection = pyplumio.connection.SerialConnection(
        device=DEVICE, timeout=10, reconnect_on_failure=False, protocol=mock_protocol
    )
    await serial_connection.connect()
    serial_asyncio_open_serial_connection.assert_called_once_with(
        url=DEVICE,
        baudrate=115200,
        bytesize=EIGHTBITS,
        parity=PARITY_NONE,
        stopbits=STOPBITS_ONE,
        timeout=10,
    )

    # Raise custom exception on connection failure.
    serial_asyncio_open_serial_connection.side_effect = SerialException
    with pytest.raises(ConnectionFailedError):
        await serial_connection.connect()


@pytest.mark.usefixtures("asyncio_open_connection")
async def test_reconnect(
    tcp_connection: pyplumio.connection.TcpConnection, caplog
) -> None:
    """Test a reconnect logic."""
    with (
        caplog.at_level(logging.ERROR),
        patch(
            "pyplumio.connection.Connection._connect",
            side_effect=(ConnectionFailedError, None),
        ) as mock_connect,
    ):
        await tcp_connection.connect()
        await tcp_connection.wait_until_done()

    assert "Can't connect to the device" in caplog.text
    assert mock_connect.call_count == 2


@pytest.mark.usefixtures("asyncio_open_connection")
async def test_connection_lost(
    mock_protocol, tcp_connection: pyplumio.connection.TcpConnection
) -> None:
    """Test that connection lost callback calls reconnect."""
    await tcp_connection.connect()
    on_connection_lost = mock_protocol.on_connection_lost.add.call_args.args[0]
    with patch("pyplumio.connection.Connection._connect") as mock_connect:
        await on_connection_lost()
        mock_connect.assert_called_once()


async def test_reconnect_logic_selection() -> None:
    """Test reconnect logic selection."""
    connection = pyplumio.connection.TcpConnection(
        host=HOST, port=PORT, reconnect_on_failure=False
    )

    with (
        patch("pyplumio.connection.Connection._connect") as mock_connect,
        patch("pyplumio.connection.Connection._reconnect") as mock_reconnect,
    ):
        await connection.connect()

    mock_reconnect.assert_not_called()
    mock_connect.assert_called_once()


@patch("pyplumio.connection.Connection.close")
@patch("pyplumio.connection.Connection.connect")
async def test_context_manager(
    mock_connect,
    mock_close,
) -> None:
    """Test context manager integration."""
    async with pyplumio.connection.TcpConnection(host=HOST, port=PORT):
        pass

    mock_connect.assert_called_once()
    mock_close.assert_called_once()


@pytest.mark.usefixtures("asyncio_open_connection")
async def test_getattr(
    mock_protocol, tcp_connection: pyplumio.connection.TcpConnection
) -> None:
    """Test that getattr is getting proxied to the protocol."""
    mock_protocol.get_device = AsyncMock()
    await tcp_connection.connect()
    await tcp_connection.get_device("test")
    mock_protocol.get_device.assert_called_once()


@pytest.mark.usefixtures("asyncio_open_connection")
async def test_close(
    mock_protocol, tcp_connection: pyplumio.connection.TcpConnection
) -> None:
    """Test connection close."""
    await tcp_connection.connect()
    await tcp_connection.close()
    mock_protocol.shutdown.assert_called_once()


async def test_repr(
    tcp_connection: pyplumio.connection.TcpConnection,
    serial_connection: pyplumio.connection.SerialConnection,
) -> None:
    """Test serializable representation."""
    assert (
        repr(tcp_connection)
        == f"TcpConnection(host={HOST}, port={PORT}, kwargs={{'timeout': 10}})"
    )
    assert repr(serial_connection) == (
        f"SerialConnection(device={DEVICE}, baudrate=115200, "
        f"kwargs={{'timeout': 10}})"
    )



"""Contains a frame reader and writer classes."""

from __future__ import annotations

import asyncio
from asyncio import IncompleteReadError, StreamReader, StreamWriter
import logging
from typing import Final, NamedTuple

from pyplumio.const import DeviceType
from pyplumio.devices import is_known_device_type
from pyplumio.exceptions import ChecksumError, ReadError, UnknownDeviceError
from pyplumio.frames import FRAME_START, Frame, bcc, struct_header
from pyplumio.helpers.timeout import timeout

READER_TIMEOUT: Final = 10
WRITER_TIMEOUT: Final = 10

MIN_FRAME_LENGTH: Final = 10
MAX_FRAME_LENGTH: Final = 1000

_LOGGER = logging.getLogger(__name__)


class FrameWriter:
    """Represents a frame writer."""

    __slots__ = ("_writer",)

    _writer: StreamWriter

    def __init__(self, writer: StreamWriter):
        """Initialize a new frame writer."""
        self._writer = writer

    @timeout(WRITER_TIMEOUT)
    async def write(self, frame: Frame) -> None:
        """Send the frame and wait until send buffer is empty."""
        self._writer.write(frame.bytes)
        await self._writer.drain()
        _LOGGER.debug("Sent frame: %s", frame)

    async def close(self) -> None:
        """Close the frame writer."""
        try:
            self._writer.close()
            await self.wait_closed()
        except (OSError, asyncio.TimeoutError):
            _LOGGER.exception("Unexpected error while closing the writer")

    @timeout(WRITER_TIMEOUT)
    async def wait_closed(self) -> None:
        """Wait until the frame writer is closed."""
        await self._writer.wait_closed()


class Header(NamedTuple):
    """Represents a frame header."""

    bytes: bytes
    frame_start: int
    frame_length: int
    recipient: int
    sender: int
    econet_type: int
    econet_version: int


class FrameReader:
    """Represents a frame reader."""

    __slots__ = ("_reader",)

    _reader: StreamReader

    def __init__(self, reader: StreamReader):
        """Initialize a new frame reader."""
        self._reader = reader

    async def _read_header(self) -> Header:
        """Locate and read a frame header.

        Raise pyplumio.ReadError if header size is too small and
        OSError if serial connection is broken.
        """
        while buffer := await self._reader.read(1):
            if FRAME_START not in buffer:
                continue

            buffer += await self._reader.read(struct_header.size - 1)
            if len(buffer) < struct_header.size:
                raise ReadError(f"Header can't be less than {struct_header.size} bytes")

            return Header(buffer, *struct_header.unpack_from(buffer))

        raise OSError("Serial connection broken")

    @timeout(READER_TIMEOUT)
    async def read(self) -> Frame | None:
        """Read the frame and return corresponding handler object.

        Raise pyplumio.ReadError on unexpected frame length or
        incomplete frame and pyplumio. Raise ChecksumError on incorrect
        frame checksum.
        """
        (
            header_bytes,
            _,
            frame_length,
            recipient,
            sender,
            econet_type,
            econet_version,
        ) = await self._read_header()

        if recipient not in (DeviceType.ECONET, DeviceType.ALL):
            return None

        if not is_known_device_type(sender):
            raise UnknownDeviceError(f"Unknown sender type ({sender})")

        if frame_length > MAX_FRAME_LENGTH or frame_length < MIN_FRAME_LENGTH:
            raise ReadError(f"Unexpected frame length ({frame_length})")

        try:
            payload = await self._reader.readexactly(frame_length - struct_header.size)
        except IncompleteReadError as e:
            raise ReadError(
                "Got an incomplete frame while trying to read "
                + f"'{frame_length - struct_header.size}' bytes"
            ) from e

        if (checksum := bcc(header_bytes + payload[:-2])) and checksum != payload[-2]:
            raise ChecksumError(
                f"Incorrect frame checksum ({checksum} != {payload[-2]})"
            )

        frame = await Frame.create(
            frame_type=payload[0],
            recipient=DeviceType(recipient),
            sender=DeviceType(sender),
            econet_type=econet_type,
            econet_version=econet_version,
            message=payload[1:-2],
        )
        _LOGGER.debug("Received frame: %s", frame)

        return frame



"""Contains an event manager class."""

from __future__ import annotations

import asyncio
from collections.abc import Callable, Coroutine
from typing import Any, Generic, TypeVar, overload

from typing_extensions import TypeAlias

from pyplumio.helpers.task_manager import TaskManager

Callback: TypeAlias = Callable[[Any], Coroutine[Any, Any, Any]]
CallbackT = TypeVar("CallbackT", bound=Callback)
T = TypeVar("T")


class EventManager(TaskManager, Generic[T]):
    """Represents an event manager."""

    __slots__ = ("data", "_events", "_callbacks")

    data: dict[str, T]
    _events: dict[str, asyncio.Event]
    _callbacks: dict[str, list[Callback]]

    def __init__(self) -> None:
        """Initialize a new event manager."""
        super().__init__()
        self.data = {}
        self._events = {}
        self._callbacks = {}

    def __getattr__(self, name: str) -> T:
        """Return attributes from the underlying data dictionary."""
        try:
            return self.data[name]
        except KeyError as e:
            raise AttributeError from e

    async def wait_for(self, name: str, timeout: float | None = None) -> None:
        """Wait for the value to become available.

        :param name: Event name or ID
        :type name: str
        :param timeout: Wait this amount of seconds for a data to
            become available, defaults to `None`
        :type timeout: float, optional
        :raise asyncio.TimeoutError: when waiting past specified timeout
        """
        if name not in self.data:
            await asyncio.wait_for(self.create_event(name).wait(), timeout=timeout)

    async def get(self, name: str, timeout: float | None = None) -> T:
        """Get the value by name.

        :param name: Event name or ID
        :type name: str
        :param timeout: Wait this amount of seconds for a data to
            become available, defaults to `None`
        :type timeout: float, optional
        :return: An event data
        :raise asyncio.TimeoutError: when waiting past specified timeout
        """
        await self.wait_for(name, timeout=timeout)
        return self.data[name]

    @overload
    def get_nowait(self, name: str, default: None = ...) -> T | None: ...

    @overload
    def get_nowait(self, name: str, default: T) -> T: ...

    def get_nowait(self, name: str, default: Any = None) -> Any:
        """Get the value by name without waiting.

        If value is not available, default value will be
        returned instead.

        :param name: Event name or ID
        :type name: str
        :param default: default value to return if data is unavailable,
            defaults to `None`
        :type default: Any, optional
        :return: An event data
        """
        try:
            return self.data[name]
        except KeyError:
            return default

    def subscribe(self, name: str, callback: CallbackT) -> CallbackT:
        """Subscribe a callback to the event.

        :param name: Event name or ID
        :type name: str
        :param callback: A coroutine callback function, that will be
            awaited on the with the event data as an argument.
        :type callback: Callback
        :return: A reference to the callback, that can be used
            with `EventManager.unsubscribe()`.
        :rtype: Callback
        """
        callbacks = self._callbacks.setdefault(name, [])
        callbacks.append(callback)
        return callback

    def subscribe_once(self, name: str, callback: Callback) -> Callback:
        """Subscribe a callback to the event once.

        Callback will be unsubscribed after single event.

        :param name: Event name or ID
        :type name: str
        :param callback: A coroutine callback function, that will be
            awaited on the with the event data as an argument.
        :type callback: Callback
        :return: A reference to the callback, that can be used
            with `EventManager.unsubscribe()`.
        :rtype: Callback
        """

        async def _call_once(value: Any) -> Any:
            """Unsubscribe callback from the event and calls it."""
            self.unsubscribe(name, _call_once)
            return await callback(value)

        return self.subscribe(name, _call_once)

    def unsubscribe(self, name: str, callback: Callback) -> bool:
        """Usubscribe a callback from the event.

        :param name: Event name or ID
        :type name: str
        :param callback: A coroutine callback function, previously
            subscribed to an event using ``subscribe()`` or
            ``subscribe_once()`` methods.
        :type callback: Callback
        :return: `True` if callback is found, `False` otherwise.
        :rtype: bool
        """
        if name in self._callbacks and callback in self._callbacks[name]:
            self._callbacks[name].remove(callback)
            return True

        return False

    async def dispatch(self, name: str, value: T) -> None:
        """Call registered callbacks and dispatch the event."""
        if callbacks := self._callbacks.get(name, None):
            for callback in list(callbacks):
                result = await callback(value)
                value = result if result is not None else value

        self.data[name] = value
        self.set_event(name)

    def dispatch_nowait(self, name: str, value: T) -> None:
        """Call a registered callbacks and dispatch the event without waiting."""
        self.create_task(self.dispatch(name, value))

    async def load(self, data: dict[str, T]) -> None:
        """Load event data."""
        await asyncio.gather(
            *(self.dispatch(name, value) for name, value in data.items())
        )

    def load_nowait(self, data: dict[str, T]) -> None:
        """Load event data without waiting."""
        self.create_task(self.load(data))

    def create_event(self, name: str) -> asyncio.Event:
        """Create an event."""
        if name in self.events:
            return self.events[name]

        event = asyncio.Event()
        self._events[name] = event
        return event

    def set_event(self, name: str) -> None:
        """Set an event."""
        if name in self.events:
            event = self.events[name]
            if not event.is_set():
                event.set()

    @property
    def events(self) -> dict[str, asyncio.Event]:
        """Return the events."""
        return self._events



"""Contains a task manager class."""

from __future__ import annotations

import asyncio
from collections.abc import Coroutine
from typing import Any


class TaskManager:
    """Represents a task manager."""

    __slots__ = ("_tasks",)

    _tasks: set[asyncio.Task]

    def __init__(self) -> None:
        """Initialize a new task manager."""
        super().__init__()
        self._tasks = set()

    def create_task(
        self, coro: Coroutine[Any, Any, Any], name: str | None = None
    ) -> asyncio.Task:
        """Create asyncio task and store a reference for it."""
        task = asyncio.create_task(coro, name=name)
        self._tasks.add(task)
        task.add_done_callback(self._tasks.discard)
        return task

    def cancel_tasks(self) -> bool:
        """Cancel all tasks."""
        return all(task.cancel() for task in self._tasks)

    async def wait_until_done(self, return_exceptions: bool = True) -> None:
        """Wait for all tasks to complete."""
        await asyncio.gather(*self._tasks, return_exceptions=return_exceptions)

    @property
    def tasks(self) -> set[asyncio.Task]:
        """Return the tasks."""
        return self._tasks



"""Contains tests for the device handler classes."""

import asyncio
from unittest.mock import AsyncMock, Mock, call, patch

import pytest

from pyplumio.const import (
    ATTR_DEVICE_INDEX,
    ATTR_FRAME_ERRORS,
    ATTR_INDEX,
    ATTR_LOADED,
    ATTR_OFFSET,
    ATTR_PARAMETER,
    ATTR_SCHEDULE,
    ATTR_SIZE,
    ATTR_STATE,
    ATTR_SWITCH,
    ATTR_TYPE,
    ATTR_VALUE,
    STATE_OFF,
    STATE_ON,
    DeviceState,
    DeviceType,
    FrameType,
    UnitOfMeasurement,
)
from pyplumio.devices import get_device_handler
from pyplumio.devices.ecomax import (
    ATTR_FUEL_BURNED,
    ATTR_MIXERS,
    ATTR_THERMOSTATS,
    SETUP_FRAME_TYPES,
    EcoMAX,
)
from pyplumio.devices.ecoster import EcoSTER
from pyplumio.devices.mixer import Mixer
from pyplumio.devices.thermostat import Thermostat
from pyplumio.exceptions import UnknownDeviceError
from pyplumio.frames import Response
from pyplumio.frames.messages import RegulatorDataMessage, SensorDataMessage
from pyplumio.frames.requests import (
    AlertsRequest,
    EcomaxControlRequest,
    RegulatorDataSchemaRequest,
    SchedulesRequest,
    SetEcomaxParameterRequest,
    SetMixerParameterRequest,
    SetScheduleRequest,
    SetThermostatParameterRequest,
    UIDRequest,
)
from pyplumio.frames.responses import (
    EcomaxParametersResponse,
    MixerParametersResponse,
    RegulatorDataSchemaResponse,
    SchedulesResponse,
    ThermostatParametersResponse,
)
from pyplumio.helpers.parameter import SET_RETRIES
from pyplumio.helpers.schedule import Schedule
from pyplumio.structures.ecomax_parameters import (
    ATTR_ECOMAX_CONTROL,
    ECOMAX_PARAMETERS,
    EcomaxNumber,
    EcomaxSwitch,
)
from pyplumio.structures.frame_versions import ATTR_FRAME_VERSIONS
from pyplumio.structures.fuel_consumption import ATTR_FUEL_CONSUMPTION
from pyplumio.structures.mixer_parameters import (
    ATTR_MIXER_PARAMETERS,
    MixerNumber,
    MixerSwitch,
)
from pyplumio.structures.mixer_sensors import ATTR_MIXER_SENSORS
from pyplumio.structures.network_info import NetworkInfo
from pyplumio.structures.regulator_data import ATTR_REGDATA
from pyplumio.structures.schedules import (
    ATTR_SCHEDULE_PARAMETER,
    ATTR_SCHEDULE_SWITCH,
    ATTR_SCHEDULES,
    ScheduleNumber,
    ScheduleSwitch,
)
from pyplumio.structures.thermostat_parameters import (
    ATTR_THERMOSTAT_PARAMETERS,
    ATTR_THERMOSTAT_PROFILE,
    ThermostatNumber,
)
from pyplumio.structures.thermostat_sensors import (
    ATTR_THERMOSTAT_SENSORS,
    ATTR_THERMOSTATS_AVAILABLE,
    ATTR_THERMOSTATS_CONNECTED,
)
from tests import load_json_test_data

UNKNOWN_DEVICE: int = 99
UNKNOWN_FRAME: int = 99


@pytest.fixture(name="ecoster")
def fixture_ecoster() -> EcoSTER:
    """Return instance of ecoster."""
    return EcoSTER(asyncio.Queue(), network=NetworkInfo())


def test_device_handler() -> None:
    """Test getting device handler class by device address."""
    handler = get_device_handler(DeviceType.ECOMAX)
    assert handler == "devices.ecomax.EcoMAX"
    with pytest.raises(UnknownDeviceError):
        handler = get_device_handler(UNKNOWN_DEVICE)


def test_ecoster(ecoster: EcoSTER) -> None:
    """Test ecoster instance."""
    assert isinstance(ecoster, EcoSTER)


async def test_int(ecoster: EcoSTER) -> None:
    """Test getting addressable device's address via int()."""
    assert int(ecoster) == DeviceType.ECOSTER


async def test_async_setup() -> None:
    """Test requesting initial data frames."""
    ecomax = EcoMAX(asyncio.Queue(), network=NetworkInfo())

    with (
        patch("pyplumio.devices.ecomax.EcoMAX.wait_for"),
        patch(
            "pyplumio.devices.ecomax.EcoMAX.request",
            side_effect=(True, True, True, True, True, True, True, True),
        ) as mock_request,
    ):
        await ecomax.async_setup()
        await ecomax.wait_until_done()

    assert await ecomax.get(ATTR_LOADED)
    assert not ecomax.data[ATTR_FRAME_ERRORS]
    assert mock_request.await_count == len(SETUP_FRAME_TYPES)


async def test_async_setup_error() -> None:
    """Test with error during requesting initial data frames."""
    ecomax = EcoMAX(asyncio.Queue(), network=NetworkInfo())

    with (
        patch("pyplumio.devices.ecomax.EcoMAX.wait_for"),
        patch(
            "pyplumio.devices.ecomax.EcoMAX.request",
            side_effect=(
                ValueError("test", FrameType.REQUEST_ALERTS),
                True,
                True,
                True,
                True,
                True,
                True,
                True,
            ),
        ) as mock_request,
    ):
        await ecomax.async_setup()
        await ecomax.wait_until_done()

    assert await ecomax.get(ATTR_LOADED)
    assert ecomax.data[ATTR_FRAME_ERRORS][0] == FrameType.REQUEST_ALERTS
    assert mock_request.await_count == len(SETUP_FRAME_TYPES)


async def test_frame_versions_update(ecomax: EcoMAX) -> None:
    """Test requesting updated frames."""
    test_data = load_json_test_data("messages/sensor_data.json")[0]
    with patch("asyncio.Queue.put_nowait") as mock_put_nowait:
        ecomax.handle_frame(SensorDataMessage(message=test_data["message"]))
        await ecomax.wait_until_done()

    mock_put_nowait.assert_has_calls(
        [
            call(RegulatorDataSchemaRequest(recipient=DeviceType.ECOMAX)),
            call(SchedulesRequest(recipient=DeviceType.ECOMAX)),
            call(UIDRequest(recipient=DeviceType.ECOMAX)),
            call(AlertsRequest(recipient=DeviceType.ECOMAX)),
        ]
    )


async def test_ecomax_data_callbacks(ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on data frames."""
    test_data = load_json_test_data("messages/sensor_data.json")[0]
    ecomax.handle_frame(SensorDataMessage(message=test_data["message"]))
    await ecomax.wait_until_done()
    heating_target = await ecomax.get("heating_target")
    assert heating_target == 41.0

    ecomax_control = await ecomax.get(ATTR_ECOMAX_CONTROL)
    assert isinstance(ecomax_control, EcomaxSwitch)
    assert ecomax_control.value == STATE_OFF

    ecomax_control_request = await ecomax_control.create_request()
    assert isinstance(ecomax_control_request, EcomaxControlRequest)
    assert ecomax_control_request.data == {ATTR_VALUE: 0}

    # Check ecomax_control reference equility.
    ecomax.handle_frame(SensorDataMessage(data={ATTR_STATE: DeviceState.WORKING}))
    await ecomax.wait_until_done()
    ecomax_control2 = await ecomax.get(ATTR_ECOMAX_CONTROL)
    assert ecomax_control2 is ecomax_control
    assert ecomax_control2.value == STATE_ON
    ecomax_control2_request = await ecomax_control.create_request()
    assert ecomax_control2_request.data == {ATTR_VALUE: 1}


async def test_ecomax_naming_collisions(ecomax: EcoMAX) -> None:
    """Test that ecoMAX sensors don't collide with ecoMAX parameters."""
    test_data = load_json_test_data("messages/sensor_data.json")[0]
    ecomax.handle_frame(SensorDataMessage(message=test_data["message"]))
    await ecomax.wait_until_done()
    for descriptions in ECOMAX_PARAMETERS.values():
        collisions = [
            description.name
            for description in descriptions
            if description.name in ecomax.data
        ]
        assert not collisions, f"found collisions: {collisions}"


async def test_ecomax_parameters_callbacks(ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on parameter frames."""
    test_data = load_json_test_data("responses/ecomax_parameters.json")[0]
    ecomax.handle_frame(EcomaxParametersResponse(message=test_data["message"]))
    await ecomax.wait_until_done()
    fuzzy_logic = await ecomax.get("fuzzy_logic")
    assert isinstance(fuzzy_logic, EcomaxSwitch)
    assert fuzzy_logic == STATE_ON
    assert fuzzy_logic.value == STATE_ON
    fuzzy_logic_request = await fuzzy_logic.create_request()
    assert isinstance(fuzzy_logic_request, SetEcomaxParameterRequest)
    assert fuzzy_logic_request.data == {
        ATTR_INDEX: 18,
        ATTR_VALUE: 1,
    }

    # Test that parameter instance is not recreated on subsequent calls.
    ecomax.handle_frame(EcomaxParametersResponse(message=test_data["message"]))
    await ecomax.wait_until_done()
    assert await ecomax.get("fuzzy_logic") is fuzzy_logic

    # Test parameter with the multiplier (heating_heat_curve)
    heating_heat_curve = await ecomax.get("heating_curve")
    assert heating_heat_curve.value == 1.3
    assert heating_heat_curve.min_value == 0.1
    assert heating_heat_curve.max_value == 4.0

    # Test setting parameter with the multiplier.
    with patch(
        "pyplumio.structures.ecomax_parameters.Parameter.set", new_callable=AsyncMock
    ) as mock_set:
        await heating_heat_curve.set(2.5)

    mock_set.assert_awaited_once_with(25, SET_RETRIES)

    # Test parameter with the offset (heating_heat_curve_shift)
    heating_heat_curve_shift = await ecomax.get("heating_curve_shift")
    assert isinstance(heating_heat_curve_shift, EcomaxNumber)
    assert heating_heat_curve_shift.value == 0.0
    assert heating_heat_curve_shift.min_value == -20.0
    assert heating_heat_curve_shift.max_value == 20.0
    assert heating_heat_curve_shift.unit_of_measurement == UnitOfMeasurement.CELSIUS

    # Test setting the parameter with the offset.
    with patch(
        "pyplumio.structures.ecomax_parameters.Parameter.set", new_callable=AsyncMock
    ) as mock_set:
        await heating_heat_curve_shift.set(1)

    mock_set.assert_awaited_once_with(21, SET_RETRIES)


async def test_unknown_ecomax_parameter(ecomax: EcoMAX, caplog) -> None:
    """Test unknown ecoMAX parameter."""
    test_data = load_json_test_data("unknown/unknown_ecomax_parameter.json")
    ecomax.handle_frame(EcomaxParametersResponse(message=test_data["message"]))
    await ecomax.wait_until_done()
    assert "unknown ecoMAX parameter (139)" in caplog.text
    assert "ParameterValues(value=1, min_value=1, max_value=1)" in caplog.text
    assert "ecoMAX 350P2-ZF" in caplog.text


@patch(
    "time.perf_counter_ns",
    side_effect=(0, 10 * 1000000000, 310 * 1000000000, 320 * 1000000000),
)
async def test_fuel_consumption_callbacks(mock_time, caplog) -> None:
    """Test callbacks dispatched on fuel consumption."""
    ecomax = EcoMAX(asyncio.Queue(), network=NetworkInfo())
    ecomax.handle_frame(Response(data={ATTR_FUEL_CONSUMPTION: 3.6}))
    await ecomax.wait_until_done()
    fuel_burned = await ecomax.get(ATTR_FUEL_BURNED)
    assert fuel_burned == 0.01
    ecomax.handle_frame(Response(data={ATTR_FUEL_CONSUMPTION: 1}))
    await ecomax.wait_until_done()
    fuel_burned = await ecomax.get(ATTR_FUEL_BURNED)
    assert "Skipping outdated fuel consumption" in caplog.text
    caplog.clear()
    ecomax.handle_frame(Response(data={ATTR_FUEL_CONSUMPTION: 7.2}))
    await ecomax.wait_until_done()
    assert "Skipping outdated fuel consumption" not in caplog.text
    fuel_burned = await ecomax.get(ATTR_FUEL_BURNED)
    assert fuel_burned == 0.02


async def test_regdata_callbacks(ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on regdata."""
    test_schema, test_regdata = (
        load_json_test_data("responses/regulator_data_schema.json")[0],
        load_json_test_data("messages/regulator_data.json")[0],
    )
    ecomax.handle_frame(RegulatorDataSchemaResponse(message=test_schema["message"]))
    await ecomax.wait_until_done()
    ecomax.handle_frame(RegulatorDataMessage(message=test_regdata["message"]))
    await ecomax.wait_until_done()
    assert await ecomax.get(ATTR_REGDATA) == test_regdata["data"][ATTR_REGDATA]
    assert (
        await ecomax.get(ATTR_FRAME_VERSIONS)
        == test_regdata["data"][ATTR_FRAME_VERSIONS]
    )


async def test_regdata_callbacks_without_schema(ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on regdata without regdata schema."""
    test_data = load_json_test_data("messages/regulator_data.json")[0]
    ecomax.handle_frame(RegulatorDataMessage(message=test_data["message"]))
    await ecomax.wait_until_done()
    assert ATTR_FRAME_VERSIONS in ecomax.data
    assert ATTR_REGDATA not in ecomax.data


async def test_mixer_sensors_callbacks(ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on mixer sensors info."""
    ecomax.handle_frame(
        SensorDataMessage(
            message=load_json_test_data("messages/sensor_data.json")[0]["message"]
        )
    )
    await ecomax.wait_until_done()
    mixers = await ecomax.get(ATTR_MIXERS)
    assert len(mixers) == 1
    mixer = mixers[4]
    assert isinstance(mixer, Mixer)
    assert mixer.index == 4
    assert mixer.data == {
        "current_temp": 20.0,
        "target_temp": 40,
        "pump": False,
        "mixer_sensors": True,
    }


async def test_mixer_sensors_callbacks_without_mixers(ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on mixer sensors without any mixers."""
    ecomax.handle_frame(MixerParametersResponse(data={ATTR_MIXER_SENSORS: {}}))
    await ecomax.wait_until_done()
    assert not await ecomax.get(ATTR_MIXER_SENSORS)


async def test_thermostat_sensors_callbacks(ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on thermostat sensors info."""
    test_data = load_json_test_data("messages/sensor_data.json")[0]
    ecomax.handle_frame(SensorDataMessage(message=test_data["message"]))
    await ecomax.wait_until_done()
    thermostats = await ecomax.get(ATTR_THERMOSTATS)
    assert len(thermostats) == 1
    thermostat = thermostats[0]
    assert isinstance(thermostat, Thermostat)
    assert thermostat.index == 0
    assert thermostat.data == {
        "state": 3,
        "current_temp": 43.5,
        "target_temp": 50.0,
        "contacts": True,
        "schedule": False,
        "thermostat_sensors": True,
    }
    assert await ecomax.get(ATTR_THERMOSTATS_AVAILABLE) == 2
    assert await ecomax.get(ATTR_THERMOSTATS_CONNECTED) == 1


async def test_thermostat_sensors_callbacks_without_thermostats(ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on thermostats sensors without any thermostats."""
    ecomax.handle_frame(
        ThermostatParametersResponse(data={ATTR_THERMOSTAT_SENSORS: {}})
    )
    await ecomax.wait_until_done()
    assert not await ecomax.get(ATTR_THERMOSTAT_SENSORS)


async def test_thermostat_parameters_callbacks(ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on thermostat parameters."""
    test_data = load_json_test_data("responses/thermostat_parameters.json")[0]
    ecomax.handle_frame(Response(data={ATTR_THERMOSTATS_AVAILABLE: 3}))
    await ecomax.wait_until_done()
    ecomax.handle_frame(ThermostatParametersResponse(message=test_data["message"]))
    await ecomax.wait_until_done()
    thermostats = await ecomax.get(ATTR_THERMOSTATS)
    assert len(thermostats) == 1
    thermostat = thermostats[0]
    assert len(thermostat.data) == 13
    party_target_temp = await thermostat.get("party_target_temp")
    assert isinstance(party_target_temp, ThermostatNumber)
    assert isinstance(party_target_temp.device, Thermostat)
    assert party_target_temp.value == 22.0
    assert party_target_temp.min_value == 10.0
    assert party_target_temp.max_value == 35.0
    assert party_target_temp.offset == 0
    party_target_temp_request = await party_target_temp.create_request()
    assert isinstance(party_target_temp_request, SetThermostatParameterRequest)
    assert party_target_temp_request.data == {
        ATTR_INDEX: 2,
        ATTR_VALUE: 220,
        ATTR_OFFSET: 0,
        ATTR_SIZE: 2,
    }

    # Test that parameter instance is not recreated on subsequent calls.
    ecomax.handle_frame(ThermostatParametersResponse(message=test_data["message"]))
    await ecomax.wait_until_done()
    thermostats = await ecomax.get(ATTR_THERMOSTATS)
    thermostat = thermostats[0]
    assert await thermostat.get("party_target_temp") is party_target_temp


async def test_thermostat_parameters_callbacks_without_thermostats(
    ecomax: EcoMAX,
) -> None:
    """Test callbacks dispatched on thermostat parameters without any thermostats."""
    test_data = load_json_test_data("responses/thermostat_parameters.json")[0]
    ecomax.handle_frame(Response(data={ATTR_THERMOSTATS_AVAILABLE: 0}))
    ecomax.handle_frame(ThermostatParametersResponse(message=test_data["message"]))
    await ecomax.wait_until_done()
    assert not await ecomax.get(ATTR_THERMOSTAT_PARAMETERS)


async def test_thermostat_profile_callbacks(ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on thermostat profile."""
    ecomax.handle_frame(Response(data={ATTR_THERMOSTATS_AVAILABLE: 3}))
    await ecomax.wait_until_done()
    ecomax.handle_frame(
        ThermostatParametersResponse(
            message=load_json_test_data("responses/thermostat_parameters.json")[0][
                "message"
            ]
        )
    )
    await ecomax.wait_until_done()
    thermostat_profile = await ecomax.get(ATTR_THERMOSTAT_PROFILE)
    assert isinstance(thermostat_profile, EcomaxNumber)
    assert thermostat_profile.value == 0.0
    assert thermostat_profile.min_value == 0.0
    assert thermostat_profile.max_value == 5.0
    thermostat_profile_request = await thermostat_profile.create_request()
    assert isinstance(thermostat_profile_request, SetThermostatParameterRequest)
    assert thermostat_profile_request.data == {
        ATTR_INDEX: 0,
        ATTR_VALUE: 0,
        ATTR_OFFSET: 0,
        ATTR_SIZE: 1,
    }

    # Test when thermostat profile is none.
    ecomax.handle_frame(Response(data={ATTR_THERMOSTAT_PROFILE: None}))
    await ecomax.wait_until_done()
    assert await ecomax.get(ATTR_THERMOSTAT_PROFILE) is None


async def test_mixer_parameters_callbacks(ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on mixer parameters."""
    test_data = load_json_test_data("responses/mixer_parameters.json")[0]
    ecomax.handle_frame(MixerParametersResponse(message=test_data["message"]))
    await ecomax.wait_until_done()
    mixers = await ecomax.get(ATTR_MIXERS)
    assert len(mixers) == 1
    mixer = mixers[0]
    assert len(mixer.data) == 7
    mixer_target_temp = await mixer.get("mixer_target_temp")
    assert isinstance(mixer_target_temp, MixerNumber)
    assert isinstance(mixer_target_temp.device, Mixer)
    assert mixer_target_temp.value == 40.0
    assert mixer_target_temp.min_value == 30.0
    assert mixer_target_temp.max_value == 60.0

    # Test creating a request.
    mixer_target_temp_request = await mixer_target_temp.create_request()
    assert isinstance(mixer_target_temp_request, SetMixerParameterRequest)
    assert mixer_target_temp_request.data == {
        ATTR_INDEX: 0,
        ATTR_VALUE: 40,
        ATTR_DEVICE_INDEX: 0,
    }

    weather_control = await mixer.get("weather_control")
    assert isinstance(weather_control, MixerSwitch)
    assert weather_control.value == STATE_ON

    heat_curve = await mixer.get("heating_curve")
    assert isinstance(heat_curve, MixerNumber)
    assert heat_curve.value == 1.3
    assert heat_curve.min_value == 1.0
    assert heat_curve.max_value == 3.0

    # Test that parameter instance is not recreated on subsequent calls.
    ecomax.handle_frame(MixerParametersResponse(message=test_data["message"]))
    await ecomax.wait_until_done()
    mixers = await ecomax.get(ATTR_MIXERS)
    mixer = mixers[0]
    assert await mixer.get("mixer_target_temp") is mixer_target_temp


async def test_mixer_parameters_callbacks_without_mixers(ecomax: EcoMAX) -> None:
    """Test mixer parameters callbacks without any mixers."""
    ecomax.handle_frame(MixerParametersResponse(data={ATTR_MIXER_PARAMETERS: {}}))
    await ecomax.wait_until_done()
    assert not await ecomax.get(ATTR_MIXER_PARAMETERS)


async def test_unknown_mixer_parameter(ecomax: EcoMAX, caplog) -> None:
    """Test unknown mixer parameter."""
    test_data = load_json_test_data("unknown/unknown_mixer_parameter.json")
    ecomax.handle_frame(MixerParametersResponse(message=test_data["message"]))
    await ecomax.wait_until_done()
    assert "unknown mixer parameter (14)" in caplog.text
    assert "ParameterValues(value=1, min_value=1, max_value=1)" in caplog.text
    assert "ecoMAX 350P2-ZF" in caplog.text


async def test_schedule_callback(ecomax: EcoMAX) -> None:
    """Test callbacks dispatched on schedule data."""
    test_data = load_json_test_data("responses/schedules.json")[0]
    ecomax.handle_frame(SchedulesResponse(message=test_data["message"]))
    await ecomax.wait_until_done()
    schedules = await ecomax.get(ATTR_SCHEDULES)
    assert len(schedules) == 2
    heating_schedule = schedules["heating"]
    assert isinstance(heating_schedule, Schedule)

    heating_schedule_switch = await ecomax.get("heating_schedule_switch")
    assert heating_schedule_switch.value == STATE_OFF
    assert isinstance(heating_schedule_switch, ScheduleSwitch)

    water_heater_schedule_parameter = await ecomax.get(
        "water_heater_schedule_parameter"
    )
    assert isinstance(water_heater_schedule_parameter, ScheduleNumber)
    assert water_heater_schedule_parameter.value == 5
    assert water_heater_schedule_parameter.min_value == 0
    assert water_heater_schedule_parameter.max_value == 30
    water_heater_schedule_parameter_request = (
        await water_heater_schedule_parameter.create_request()
    )
    assert isinstance(water_heater_schedule_parameter_request, SetScheduleRequest)
    assert water_heater_schedule_parameter_request.data == {
        ATTR_TYPE: "water_heater",
        ATTR_SWITCH: ecomax.data[f"water_heater_{ATTR_SCHEDULE_SWITCH}"],
        ATTR_PARAMETER: ecomax.data[f"water_heater_{ATTR_SCHEDULE_PARAMETER}"],
        ATTR_SCHEDULE: ecomax.data[ATTR_SCHEDULES]["water_heater"],
    }

    schedule_data = test_data["data"][ATTR_SCHEDULES][0][1]
    for index, weekday in enumerate(
        ("monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday")
    ):
        schedule = getattr(heating_schedule, weekday)
        assert schedule.intervals == schedule_data[index]

    # Test that parameter instance is not recreated on subsequent calls.
    ecomax.handle_frame(SchedulesResponse(message=test_data["message"]))
    await ecomax.wait_until_done()
    assert await ecomax.get("heating_schedule_switch") is heating_schedule_switch


async def test_request(ecomax: EcoMAX) -> None:
    """Test requesting the value."""
    with (
        patch(
            "pyplumio.devices.ecomax.EcoMAX.get",
            side_effect=(asyncio.TimeoutError, True),
        ) as mock_get,
        patch("asyncio.Queue.put_nowait") as mock_put_nowait,
    ):
        assert await ecomax.request("foo", FrameType.REQUEST_ALERTS, timeout=5)

    mock_get.assert_awaited_with("foo", timeout=5)
    assert mock_put_nowait.call_count == 2
    args, _ = mock_put_nowait.call_args
    assert isinstance(args[0], AlertsRequest)


async def test_request_error(ecomax: EcoMAX) -> None:
    """Test requesting the value with error."""
    with (
        patch(
            "pyplumio.devices.ecomax.EcoMAX.get",
            side_effect=(asyncio.TimeoutError, asyncio.TimeoutError),
        ),
        pytest.raises(ValueError),
    ):
        await ecomax.request("foo", FrameType.REQUEST_ALERTS, retries=1)


@patch("pyplumio.helpers.parameter.Parameter.pending_update", False)
async def test_set(ecomax: EcoMAX) -> None:
    """Test setting parameter value via set helper."""
    test_ecomax_data, test_thermostat_data, test_mixer_data = (
        load_json_test_data("responses/ecomax_parameters.json")[0],
        load_json_test_data("responses/thermostat_parameters.json")[0],
        load_json_test_data("responses/mixer_parameters.json")[0],
    )
    ecomax.handle_frame(Response(data={ATTR_THERMOSTATS_AVAILABLE: 3}))
    await ecomax.wait_until_done()
    ecomax.handle_frame(EcomaxParametersResponse(message=test_ecomax_data["message"]))
    ecomax.handle_frame(
        ThermostatParametersResponse(message=test_thermostat_data["message"])
    )
    ecomax.handle_frame(MixerParametersResponse(message=test_mixer_data["message"]))
    await ecomax.wait_until_done()

    # Test setting an ecomax parameter.
    assert await ecomax.set("max_fuel_flow", 13.0)
    max_fuel_flow = await ecomax.get("max_fuel_flow")
    assert max_fuel_flow.value == 13.0

    # Test setting an ecomax parameter without blocking.
    with (
        patch("pyplumio.devices.Device.set", new_callable=Mock),
        patch(
            "pyplumio.helpers.task_manager.TaskManager.create_task"
        ) as mock_create_task,
    ):
        ecomax.set_nowait("max_fuel_flow", 10)

    mock_create_task.assert_called_once()

    # Test setting a thermostat parameter.
    thermostat = ecomax.data[ATTR_THERMOSTATS][0]
    assert await thermostat.set("party_target_temp", 21.0)
    target_party_temp = await thermostat.get("party_target_temp")
    assert target_party_temp.value == 21.0

    # Test setting a mixer parameter.
    mixer = ecomax.data[ATTR_MIXERS][0]
    assert await mixer.set("mixer_target_temp", 35.0)
    mixer_target_temp = await mixer.get("mixer_target_temp")
    assert mixer_target_temp.value == 35.0

    # Test with invalid parameter.
    ecomax.data["bar"] = Mock()
    with pytest.raises(TypeError):
        await ecomax.set("bar", 1)


async def test_turn_on(ecomax: EcoMAX, caplog) -> None:
    """Test turning the controller on."""
    assert not await ecomax.turn_on()
    assert "ecoMAX control isn't available" in caplog.text
    ecomax.data[ATTR_ECOMAX_CONTROL] = AsyncMock()
    assert await ecomax.turn_on()
    ecomax.data[ATTR_ECOMAX_CONTROL].turn_on.assert_awaited_once()


async def test_turn_off(ecomax: EcoMAX, caplog) -> None:
    """Test turning the controller off."""
    await ecomax.turn_off()
    assert "ecoMAX control isn't available" in caplog.text
    ecomax.data[ATTR_ECOMAX_CONTROL] = AsyncMock()
    await ecomax.turn_off()
    ecomax.data[ATTR_ECOMAX_CONTROL].turn_off.assert_awaited_once()


@patch("pyplumio.devices.ecomax.EcoMAX.create_task")
@patch("pyplumio.devices.ecomax.EcoMAX.turn_on", new_callable=Mock)
async def test_turn_on_nowait(mock_create_task, mock_turn_on, ecomax: EcoMAX) -> None:
    """Test turning the controller on without waiting."""
    ecomax.turn_on_nowait()
    await ecomax.wait_until_done()
    mock_create_task.assert_called_once()
    mock_turn_on.assert_called_once()


@patch("pyplumio.devices.ecomax.EcoMAX.create_task")
@patch("pyplumio.devices.ecomax.EcoMAX.turn_off", new_callable=Mock)
async def test_turn_off_nowait(mock_create_task, mock_turn_off, ecomax: EcoMAX) -> None:
    """Test turning the controller on without waiting."""
    ecomax.turn_off_nowait()
    await ecomax.wait_until_done()
    mock_create_task.assert_called_once()
    mock_turn_off.assert_called_once()


async def test_shutdown(ecomax: EcoMAX) -> None:
    """Test device tasks shutdown."""
    ecomax.handle_frame(
        SensorDataMessage(
            message=load_json_test_data("messages/sensor_data.json")[0]["message"]
        )
    )
    await ecomax.wait_until_done()

    with (
        patch("pyplumio.devices.mixer.Mixer.shutdown") as mock_mixer_shutdown,
        patch(
            "pyplumio.devices.thermostat.Thermostat.shutdown"
        ) as mock_thermostat_shutdown,
        patch("pyplumio.devices.Device.cancel_tasks") as mock_cancel_tasks,
        patch("pyplumio.devices.Device.wait_until_done") as mock_wait_until_done,
    ):
        await ecomax.shutdown()

    mock_wait_until_done.assert_awaited_once()
    mock_cancel_tasks.assert_called_once()
    mock_thermostat_shutdown.assert_awaited_once()
    mock_mixer_shutdown.assert_awaited_once()



"""Contains a mixer class."""

from __future__ import annotations

import asyncio
from collections.abc import Coroutine, Generator, Sequence
import logging
from typing import TYPE_CHECKING, Any

from pyplumio.devices import AddressableDevice, SubDevice
from pyplumio.helpers.parameter import ParameterValues
from pyplumio.structures.mixer_parameters import (
    ATTR_MIXER_PARAMETERS,
    MIXER_PARAMETERS,
    MixerNumber,
    MixerSwitch,
    MixerSwitchDescription,
)
from pyplumio.structures.mixer_sensors import ATTR_MIXER_SENSORS
from pyplumio.structures.product_info import ATTR_PRODUCT, ProductInfo

if TYPE_CHECKING:
    from pyplumio.frames import Frame

_LOGGER = logging.getLogger(__name__)


class Mixer(SubDevice):
    """Represents an mixer."""

    def __init__(
        self, queue: asyncio.Queue[Frame], parent: AddressableDevice, index: int = 0
    ):
        """Initialize a new mixer."""
        super().__init__(queue, parent, index)
        self.subscribe(ATTR_MIXER_SENSORS, self._handle_mixer_sensors)
        self.subscribe(ATTR_MIXER_PARAMETERS, self._handle_mixer_parameters)

    async def _handle_mixer_sensors(self, sensors: dict[str, Any]) -> bool:
        """Handle mixer sensors.

        For each sensor dispatch an event with the
        sensor's name and value.
        """
        await asyncio.gather(
            *(self.dispatch(name, value) for name, value in sensors.items())
        )
        return True

    async def _handle_mixer_parameters(
        self, parameters: Sequence[tuple[int, ParameterValues]]
    ) -> bool:
        """Handle mixer parameters.

        For each parameter dispatch an event with the
        parameter's name and value.
        """
        product: ProductInfo = await self.parent.get(ATTR_PRODUCT)

        def _mixer_parameter_events() -> Generator[Coroutine, Any, None]:
            """Get dispatch calls for mixer parameter events."""
            for index, values in parameters:
                try:
                    description = MIXER_PARAMETERS[product.type][index]
                except IndexError:
                    _LOGGER.warning(
                        (
                            "Encountered unknown mixer parameter (%i): %s. "
                            "Your device isn't fully compatible with this software and "
                            "may not work properly. "
                            "Please visit the issue tracker and open a feature "
                            "request to support %s"
                        ),
                        index,
                        values,
                        product.model,
                    )
                    return

                handler = (
                    MixerSwitch
                    if isinstance(description, MixerSwitchDescription)
                    else MixerNumber
                )
                yield self.dispatch(
                    description.name,
                    handler.create_or_update(
                        device=self,
                        description=description,
                        values=values,
                        index=index,
                    ),
                )

        await asyncio.gather(*_mixer_parameter_events())
        return True



"""Contains fixtures for the test suite."""

import asyncio
from unittest.mock import patch

import pytest

from pyplumio.const import ProductType
from pyplumio.devices.ecomax import EcoMAX
from pyplumio.structures.network_info import NetworkInfo
from pyplumio.structures.product_info import ATTR_PRODUCT, ProductInfo


@pytest.fixture(name="ecomax")
def fixture_ecomax() -> EcoMAX:
    """Return an ecoMAX object."""
    ecomax = EcoMAX(asyncio.Queue(), network=NetworkInfo())
    ecomax.data[ATTR_PRODUCT] = ProductInfo(
        type=ProductType.ECOMAX_P,
        id=90,
        uid="TEST",
        logo=23040,
        image=2816,
        model="ecoMAX 350P2-ZF",
    )
    return ecomax


@pytest.fixture(autouse=True)
def bypass_asyncio_sleep():
    """Bypass an asyncio sleep."""
    with patch("asyncio.sleep"):
        yield



"""Contains a mixer parameter structure decoder."""

from __future__ import annotations

from collections.abc import Generator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Final

from dataslots import dataslots

from pyplumio.const import (
    ATTR_DEVICE_INDEX,
    ATTR_INDEX,
    ATTR_VALUE,
    FrameType,
    ProductType,
    UnitOfMeasurement,
)
from pyplumio.frames import Request
from pyplumio.helpers.parameter import (
    SET_RETRIES,
    Number,
    NumberDescription,
    Parameter,
    ParameterDescription,
    ParameterValues,
    Switch,
    SwitchDescription,
    unpack_parameter,
)
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

if TYPE_CHECKING:
    from pyplumio.devices.mixer import Mixer

ATTR_MIXER_PARAMETERS: Final = "mixer_parameters"

MIXER_PARAMETER_SIZE: Final = 3


@dataclass
class MixerParameterDescription(ParameterDescription):
    """Represents a mixer parameter description."""

    __slots__ = ()

    multiplier: float = 1.0
    offset: int = 0


class MixerParameter(Parameter):
    """Represent a mixer parameter."""

    __slots__ = ()

    device: Mixer
    description: MixerParameterDescription

    async def create_request(self) -> Request:
        """Create a request to change the parameter."""
        return await Request.create(
            FrameType.REQUEST_SET_MIXER_PARAMETER,
            recipient=self.device.parent.address,
            data={
                ATTR_INDEX: self._index,
                ATTR_VALUE: self.values.value,
                ATTR_DEVICE_INDEX: self.device.index,
            },
        )


@dataslots
@dataclass
class MixerNumberDescription(MixerParameterDescription, NumberDescription):
    """Represent a mixer number description."""


class MixerNumber(MixerParameter, Number):
    """Represents a mixer number."""

    __slots__ = ()

    description: MixerNumberDescription

    async def set(self, value: int | float, retries: int = SET_RETRIES) -> bool:
        """Set a parameter value."""
        value = (value + self.description.offset) / self.description.multiplier
        return await super().set(value, retries)

    @property
    def value(self) -> float:
        """Return the parameter value."""
        return (
            self.values.value - self.description.offset
        ) * self.description.multiplier

    @property
    def min_value(self) -> float:
        """Return the minimum allowed value."""
        return (
            self.values.min_value - self.description.offset
        ) * self.description.multiplier

    @property
    def max_value(self) -> float:
        """Return the maximum allowed value."""
        return (
            self.values.max_value - self.description.offset
        ) * self.description.multiplier


@dataslots
@dataclass
class MixerSwitchDescription(MixerParameterDescription, SwitchDescription):
    """Represents a mixer switch description."""


class MixerSwitch(MixerParameter, Switch):
    """Represents a mixer switch."""

    __slots__ = ()

    description: MixerSwitchDescription


MIXER_PARAMETERS: dict[ProductType, tuple[MixerParameterDescription, ...]] = {
    ProductType.ECOMAX_P: (
        MixerNumberDescription(
            name="mixer_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="min_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="max_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="thermostat_decrease_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerSwitchDescription(
            name="weather_control",
        ),
        MixerNumberDescription(
            name="heating_curve",
            multiplier=0.1,
        ),
        MixerNumberDescription(
            name="heating_curve_shift",
            offset=20,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="weather_factor",
        ),
        MixerNumberDescription(
            name="work_mode",
        ),
        MixerNumberDescription(
            name="mixer_input_dead_zone",
            multiplier=0.1,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerSwitchDescription(
            name="thermostat_operation",
        ),
        MixerNumberDescription(
            name="thermostat_mode",
        ),
        MixerSwitchDescription(
            name="disable_pump_on_thermostat",
        ),
        MixerSwitchDescription(
            name="summer_work",
        ),
    ),
    ProductType.ECOMAX_I: (
        MixerNumberDescription(
            name="work_mode",
        ),
        MixerNumberDescription(
            name="circuit_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="day_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="night_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="min_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="max_target_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerSwitchDescription(
            name="summer_work",
        ),
        MixerSwitchDescription(
            name="weather_control",
        ),
        MixerNumberDescription(
            name="enable_circuit",
        ),
        MixerNumberDescription(
            name="constant_water_preset_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="thermostat_decrease_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="thermostat_correction",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerSwitchDescription(
            name="thermostat_pump_lock",
        ),
        MixerNumberDescription(
            name="valve_opening_time",
            unit_of_measurement=UnitOfMeasurement.SECONDS,
        ),
        MixerNumberDescription(
            name="mixer_input_dead_zone",
            multiplier=0.1,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="proportional_range",
        ),
        MixerNumberDescription(
            name="integration_time_constant",
        ),
        MixerNumberDescription(
            name="heating_curve",
            multiplier=0.1,
        ),
        MixerNumberDescription(
            name="heating_curve_shift",
            offset=20,
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
        MixerNumberDescription(
            name="thermostat_mode",
        ),
        MixerNumberDescription(
            name="decreasing_constant_water_temp",
            unit_of_measurement=UnitOfMeasurement.CELSIUS,
        ),
    ),
}


class MixerParametersStructure(StructureDecoder):
    """Represents a mixer parameters data structure."""

    _offset: int

    def _mixer_parameter(
        self, message: bytearray, start: int, end: int
    ) -> Generator[tuple[int, ParameterValues], None, None]:
        """Get a single mixer parameter."""
        for index in range(start, start + end):
            if parameter := unpack_parameter(message, self._offset):
                yield (index, parameter)

            self._offset += MIXER_PARAMETER_SIZE

    def _mixer_parameters(
        self, message: bytearray, mixers: int, start: int, end: int
    ) -> Generator[tuple[int, list[tuple[int, ParameterValues]]], None, None]:
        """Get parameters for a mixer."""
        for index in range(mixers):
            if parameters := list(self._mixer_parameter(message, start, end)):
                yield (index, parameters)

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        start = message[offset + 1]
        end = message[offset + 2]
        mixers = message[offset + 3]
        self._offset = offset + 4
        return (
            ensure_dict(
                data,
                {
                    ATTR_MIXER_PARAMETERS: dict(
                        self._mixer_parameters(message, mixers, start, end)
                    )
                },
            ),
            self._offset,
        )



"""Contains tests for the filter classes."""

from datetime import datetime
from unittest.mock import AsyncMock, patch

import pytest

from pyplumio.filters import aggregate, custom, debounce, delta, on_change, throttle
from pyplumio.helpers.parameter import Parameter, ParameterValues
from pyplumio.structures.alerts import Alert


async def test_on_change() -> None:
    """Test on change filter."""
    test_callback = AsyncMock()
    wrapped_callback = on_change(test_callback)
    await wrapped_callback(1)
    test_callback.assert_awaited_once_with(1)
    test_callback.reset_mock()

    # Check that callback is not awaited on insignificant change.
    await wrapped_callback(1.01)
    test_callback.assert_not_awaited()

    await wrapped_callback(1.1)
    test_callback.assert_awaited_once_with(1.1)

    # Test equality with callback function and different instance.
    assert test_callback == wrapped_callback
    assert wrapped_callback == on_change(test_callback)
    assert wrapped_callback.__eq__("you shall not pass") is NotImplemented


async def test_on_change_parameter() -> None:
    """Test on change filter with parameters."""
    test_callback = AsyncMock()
    test_parameter = AsyncMock(spec=Parameter)
    test_parameter.values = ParameterValues(0, 0, 1)
    test_parameter.pending_update = False
    wrapped_callback = on_change(test_callback)
    await wrapped_callback(test_parameter)
    test_callback.assert_awaited_once_with(test_parameter)
    test_callback.reset_mock()

    # Check that callback is not awaited with no change.
    await wrapped_callback(test_parameter)
    test_callback.assert_not_awaited()

    # Check that callback is awaited on local value change.
    test_parameter.pending_update = True
    await wrapped_callback(test_parameter)
    test_callback.assert_awaited_once_with(test_parameter)
    test_callback.reset_mock()
    test_parameter.pending_update = False

    # Check that callback is awaited on remote value change.
    test_parameter.values = ParameterValues(1, 0, 1)
    await wrapped_callback(test_parameter)
    test_callback.assert_awaited_once_with(test_parameter)
    test_callback.reset_mock()

    # Check that callback is awaited on min value change.
    test_parameter.values = ParameterValues(1, 1, 1)
    await wrapped_callback(test_parameter)
    test_callback.assert_awaited_once_with(test_parameter)
    test_callback.reset_mock()

    # Check that callback is awaited on max value change.
    test_parameter.values = ParameterValues(1, 1, 2)
    await wrapped_callback(test_parameter)
    test_callback.assert_awaited_once_with(test_parameter)


async def test_debounce() -> None:
    """Test debounce filter."""
    test_callback = AsyncMock()
    wrapped_callback = debounce(test_callback, min_calls=3)
    await wrapped_callback(1)
    test_callback.assert_awaited_once_with(1)
    test_callback.reset_mock()

    # Ignore stray "1" and only await callback on a "2".
    await wrapped_callback(2)
    await wrapped_callback(1)
    await wrapped_callback(2)
    await wrapped_callback(2)
    await wrapped_callback(2)
    test_callback.assert_awaited_once_with(2)


@patch("time.monotonic", side_effect=(0, 1, 5, 6))
async def test_throttle(mock_time) -> None:
    """Test throttle filter."""
    test_callback = AsyncMock()
    wrapped_callback = throttle(test_callback, seconds=5)
    await wrapped_callback(1)
    test_callback.assert_awaited_once_with(1)
    test_callback.reset_mock()

    # One second passed.
    await wrapped_callback(2)
    test_callback.assert_not_awaited()

    # Five seconds passed.
    await wrapped_callback(3)
    test_callback.assert_awaited_once_with(3)
    test_callback.reset_mock()

    # Six seconds passed.
    await wrapped_callback(4)
    test_callback.assert_not_awaited()


async def test_delta() -> None:
    """Test delta filter."""
    test_callback = AsyncMock()
    wrapped_callback = delta(test_callback)

    await wrapped_callback(5)
    test_callback.assert_not_awaited()

    await wrapped_callback(3)
    test_callback.assert_awaited_once_with(-2)
    test_callback.reset_mock()

    # Test with list of alerts.
    alert1 = Alert(code=0, from_dt=datetime.now(), to_dt=None)
    alert2 = Alert(code=1, from_dt=datetime.now(), to_dt=None)
    alert3 = Alert(code=2, from_dt=datetime.now(), to_dt=None)
    wrapped_callback = delta(test_callback)
    await wrapped_callback([alert1, alert2])
    await wrapped_callback([alert3, alert2])
    test_callback.assert_awaited_once_with([alert3])
    test_callback.reset_mock()

    # Test with unknown.
    wrapped_callback = delta(test_callback)
    await wrapped_callback("foo")
    await wrapped_callback("bar")
    test_callback.assert_not_awaited()


@patch("time.monotonic", side_effect=(0, 0, 1, 5, 6, 7))
async def test_aggregate(mock_time) -> None:
    """Test aggregate filter."""
    test_callback = AsyncMock()
    wrapped_callback = aggregate(test_callback, seconds=5)

    # Zero seconds passed.
    await wrapped_callback(1)
    test_callback.assert_not_awaited()

    # One second passed.
    await wrapped_callback(1)
    test_callback.assert_not_awaited()

    # Five seconds passed.
    await wrapped_callback(3)
    test_callback.assert_awaited_once_with(5)
    test_callback.reset_mock()

    # Six seconds passed.
    await wrapped_callback(3)
    test_callback.assert_not_awaited()

    # Test with non-numeric value.
    with pytest.raises(ValueError):
        await wrapped_callback("banana")


async def test_custom() -> None:
    """Test custom filter."""
    test_callback = AsyncMock()
    wrapped_callback = custom(test_callback, lambda x: len(x) == 4)

    # Test that callback is not called when a list contains 2 items.
    await wrapped_callback([1, 2])
    test_callback.assert_not_awaited()

    # Test that callback is called when a list contains 4 items.
    await wrapped_callback([1, 2, 3, 4])
    test_callback.assert_awaited_once_with([1, 2, 3, 4])
    test_callback.reset_mock()

    # Test that callback is not called when list is empty.
    await wrapped_callback([])
    test_callback.assert_not_awaited()



"""Contains a connection class."""

from __future__ import annotations

from abc import ABC, abstractmethod
import asyncio
import logging
from typing import Any, Final, cast

from serial import EIGHTBITS, PARITY_NONE, STOPBITS_ONE, SerialException

from pyplumio.exceptions import ConnectionFailedError
from pyplumio.helpers.task_manager import TaskManager
from pyplumio.helpers.timeout import timeout
from pyplumio.protocol import AsyncProtocol, Protocol

_LOGGER = logging.getLogger(__name__)

CONNECT_TIMEOUT: Final = 5
RECONNECT_TIMEOUT: Final = 20

try:
    import serial_asyncio_fast as pyserial_asyncio

    _LOGGER.info("Using pyserial-asyncio-fast in place of pyserial-asyncio")
except ImportError:
    import serial_asyncio as pyserial_asyncio


class Connection(ABC, TaskManager):
    """Represents a connection.

    All specific connection classes MUST be inherited from this class.
    """

    _protocol: Protocol
    _reconnect_on_failure: bool
    _kwargs: dict[str, Any]

    def __init__(
        self,
        protocol: Protocol | None = None,
        reconnect_on_failure: bool = True,
        **kwargs: Any,
    ) -> None:
        """Initialize a new connection."""
        super().__init__()
        if protocol is None:
            protocol = AsyncProtocol()

        if reconnect_on_failure:
            protocol.on_connection_lost.add(self._reconnect)

        self._reconnect_on_failure = reconnect_on_failure
        self._protocol = protocol
        self._kwargs = kwargs

    async def __aenter__(self) -> Connection:
        """Provide an entry point for the context manager."""
        await self.connect()
        return self

    async def __aexit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        """Provide an exit point for the context manager."""
        await self.close()

    def __getattr__(self, name: str) -> Any:
        """Return an attributes from the underlying protocol object."""
        return getattr(self.protocol, name)

    async def _connect(self) -> None:
        """Establish connection and initialize the protocol object."""
        try:
            reader, writer = await self._open_connection()
            self.protocol.connection_established(reader, writer)
        except (OSError, SerialException, asyncio.TimeoutError) as err:
            raise ConnectionFailedError from err

    async def _reconnect(self) -> None:
        """Try to connect and reconnect on failure."""
        try:
            await self._connect()
        except ConnectionFailedError:
            _LOGGER.error(
                "Can't connect to the device, retrying in %.1f seconds",
                RECONNECT_TIMEOUT,
            )
            await asyncio.sleep(RECONNECT_TIMEOUT)
            self.create_task(self._reconnect())

    async def connect(self) -> None:
        """Open the connection.

        Initialize a connection via connect or reconnect
        routines, depending on '_reconnect_on_failure' property.
        """
        await (self._reconnect if self._reconnect_on_failure else self._connect)()

    async def close(self) -> None:
        """Close the connection."""
        self.cancel_tasks()
        await self.protocol.shutdown()

    @property
    def protocol(self) -> Protocol:
        """Return the protocol object."""
        return self._protocol

    @timeout(CONNECT_TIMEOUT)
    @abstractmethod
    async def _open_connection(
        self,
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Open the connection and return reader and writer objects."""


class TcpConnection(Connection):
    """Represents a TCP connection."""

    host: str
    port: int

    def __init__(
        self,
        host: str,
        port: int,
        *,
        protocol: Protocol | None = None,
        reconnect_on_failure: bool = True,
        **kwargs: Any,
    ) -> None:
        """Initialize a new TCP connection."""
        super().__init__(protocol, reconnect_on_failure, **kwargs)
        self.host = host
        self.port = port

    def __repr__(self) -> str:
        """Return a serializable string representation."""
        return (
            f"TcpConnection(host={self.host}, port={self.port}, kwargs={self._kwargs})"
        )

    @timeout(CONNECT_TIMEOUT)
    async def _open_connection(
        self,
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Open the connection and return reader and writer objects."""
        return await asyncio.open_connection(
            host=self.host, port=self.port, **self._kwargs
        )


class SerialConnection(Connection):
    """Represents a serial connection."""

    device: str
    baudrate: int

    def __init__(
        self,
        device: str,
        baudrate: int = 115200,
        *,
        protocol: Protocol | None = None,
        reconnect_on_failure: bool = True,
        **kwargs: Any,
    ) -> None:
        """Initialize a new serial connection."""
        super().__init__(protocol, reconnect_on_failure, **kwargs)
        self.device = device
        self.baudrate = baudrate

    def __repr__(self) -> str:
        """Return a serializable string representation."""
        return (
            "SerialConnection("
            f"device={self.device}, "
            f"baudrate={self.baudrate}, "
            f"kwargs={self._kwargs})"
        )

    @timeout(CONNECT_TIMEOUT)
    async def _open_connection(
        self,
    ) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """Open the connection and return reader and writer objects."""
        return cast(
            tuple[asyncio.StreamReader, asyncio.StreamWriter],
            await pyserial_asyncio.open_serial_connection(
                url=self.device,
                baudrate=self.baudrate,
                bytesize=EIGHTBITS,
                parity=PARITY_NONE,
                stopbits=STOPBITS_ONE,
                **self._kwargs,
            ),
        )



"""Contains a schedule helper classes."""

from __future__ import annotations

from collections.abc import Iterable, Iterator, MutableMapping
from dataclasses import dataclass
import datetime as dt
from functools import lru_cache
import math
from typing import Final, Literal

from pyplumio.const import STATE_OFF, STATE_ON, FrameType
from pyplumio.devices import AddressableDevice
from pyplumio.frames import Request
from pyplumio.structures.schedules import collect_schedule_data

TIME_FORMAT: Final = "%H:%M"
START_OF_DAY: Final = "00:00"
END_OF_DAY: Final = "00:00"

STATE_NIGHT: Final = "night"
STATE_DAY: Final = "day"

ON_STATES: Final = (STATE_ON, STATE_DAY)
OFF_STATES: Final = (STATE_OFF, STATE_NIGHT)
ALLOWED_STATES: Final = ON_STATES + OFF_STATES


@lru_cache(maxsize=10)
def _parse_interval(start: str, end: str) -> tuple[int, int]:
    """Parse an interval string.

    Intervals should be specified in '%H:%M' format.
    """
    start_dt = dt.datetime.strptime(start, TIME_FORMAT)
    end_dt = dt.datetime.strptime(end, TIME_FORMAT)
    start_of_day_dt = dt.datetime.strptime(START_OF_DAY, TIME_FORMAT)
    if end_dt == start_of_day_dt:
        # Upper bound of interval is midnight.
        end_dt += dt.timedelta(hours=23, minutes=30)

    if end_dt <= start_dt:
        raise ValueError(
            f"Invalid interval ({start}, {end}). "
            + "Lower boundary must be less than upper."
        )

    start_index = math.floor((start_dt - start_of_day_dt).total_seconds() // (60 * 30))
    end_index = math.floor((end_dt - start_of_day_dt).total_seconds() // (60 * 30))

    return start_index, end_index


class ScheduleDay(MutableMapping):
    """Represents a single day of schedule."""

    __slots__ = ("_intervals",)

    _intervals: list[bool]

    def __init__(self, intervals: list[bool]):
        """Initialize a new schedule day."""
        self._intervals = intervals

    def __repr__(self) -> str:
        """Return serializable representation of the class."""
        return f"ScheduleDay({self._intervals})"

    def __len__(self) -> int:
        """Return a schedule length."""
        return len(self._intervals)

    def __iter__(self) -> Iterator[bool]:
        """Return an iterator."""
        return self._intervals.__iter__()

    def __getitem__(self, index: int) -> bool:
        """Return a schedule item."""
        return self._intervals.__getitem__(index)

    def __delitem__(self, index: int) -> None:
        """Delete a schedule item."""
        return self._intervals.__delitem__(index)

    def __setitem__(self, index: int, value: bool) -> None:
        """Set a schedule item."""
        return self._intervals.__setitem__(index, value)

    def append(self, item: bool) -> None:
        """Append a value to the interval."""
        self._intervals.append(item)

    def set_state(
        self,
        state: Literal["on", "off", "day", "night"],
        start: str = START_OF_DAY,
        end: str = END_OF_DAY,
    ) -> None:
        """Set an interval state.

        Can be on of the following:
        'on', 'off', 'day' or 'night'.
        """
        if state not in ALLOWED_STATES:
            raise ValueError(f'state "{state}" is not allowed')

        index, end_index = _parse_interval(start, end)
        while index <= end_index:
            self._intervals[index] = state in ON_STATES
            index += 1

    def set_on(self, start: str = START_OF_DAY, end: str = END_OF_DAY) -> None:
        """Set an interval state to 'on'."""
        self.set_state(STATE_ON, start, end)

    def set_off(self, start: str = START_OF_DAY, end: str = END_OF_DAY) -> None:
        """Set an interval state to 'off'."""
        self.set_state(STATE_OFF, start, end)

    @property
    def intervals(self) -> list[bool]:
        """Return the schedule intervals."""
        return self._intervals


@dataclass
class Schedule(Iterable):
    """Represents a weekly schedule."""

    __slots__ = (
        "name",
        "device",
        "monday",
        "tuesday",
        "wednesday",
        "thursday",
        "friday",
        "saturday",
        "sunday",
    )

    name: str
    device: AddressableDevice
    monday: ScheduleDay
    tuesday: ScheduleDay
    wednesday: ScheduleDay
    thursday: ScheduleDay
    friday: ScheduleDay
    saturday: ScheduleDay
    sunday: ScheduleDay

    def __iter__(self) -> Iterator[ScheduleDay]:
        """Return list of days."""
        return (
            self.sunday,
            self.monday,
            self.tuesday,
            self.wednesday,
            self.thursday,
            self.friday,
            self.saturday,
        ).__iter__()

    async def commit(self) -> None:
        """Commit a weekly schedule to the device."""
        await self.device.queue.put(
            await Request.create(
                FrameType.REQUEST_SET_SCHEDULE,
                recipient=self.device.address,
                data=collect_schedule_data(self.name, self.device),
            )
        )



"""Contains tests for the abstract frame class."""

from __future__ import annotations

from typing import ClassVar

import pytest

from pyplumio.const import DeviceType, FrameType
from pyplumio.exceptions import UnknownFrameError
from pyplumio.frames import (
    ECONET_TYPE,
    ECONET_VERSION,
    Request,
    Response,
    get_frame_handler,
    struct_header,
)
from pyplumio.frames.responses import ProgramVersionResponse
from pyplumio.structures.program_version import ATTR_VERSION, VersionInfo


class RequestFrame(Request):
    """Representation of a request frame."""

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_PROGRAM_VERSION


class ResponseFrame(Response):
    """Representation of a response frame."""

    frame_type: ClassVar[FrameType] = FrameType.RESPONSE_PROGRAM_VERSION


@pytest.fixture(name="request_frame")
def fixture_request_frame() -> Request:
    """Return a request frame object."""
    return RequestFrame(foo=0)


@pytest.fixture(name="response_frame")
def fixture_response_frame() -> Response:
    """Return a response frame object."""
    return ResponseFrame(foo=0)


@pytest.fixture(name="frames")
def fixture_frames(
    request_frame: Request, response_frame: Response
) -> tuple[Request, Response]:
    """Return both request and response frame objects as a tuple."""
    return (request_frame, response_frame)


@pytest.fixture(name="types")
def fixture_types() -> tuple[int, int]:
    """Return request and response types as a tuple."""
    return (FrameType.REQUEST_PROGRAM_VERSION, FrameType.RESPONSE_PROGRAM_VERSION)


def test_decode_create_message(frames: tuple[Request, Response]) -> None:
    """Test creating and decoding frame message."""
    for frame in frames:
        assert frame.create_message(data={}) == bytearray()
        assert frame.decode_message(message=bytearray()) == {}


def test_get_frame_handler() -> None:
    """Test getting a frame handler."""
    assert get_frame_handler(0x18) == "frames.requests.StopMasterRequest"
    with pytest.raises(UnknownFrameError):
        get_frame_handler(0x0)


def test_passing_frame_type(
    frames: tuple[Request, Response], types: tuple[int, int]
) -> None:
    """Test getting a frame type."""
    for index, frame in enumerate(frames):
        assert frame.frame_type == types[index]


def test_frame_attributes(frames: tuple[Request, Response]) -> None:
    """Test accessing frame attributes."""
    for frame in frames:
        assert frame.recipient == DeviceType.ALL
        assert frame.message == b""
        assert frame.sender == DeviceType.ECONET
        assert frame.econet_type == ECONET_TYPE
        assert frame.econet_version == ECONET_VERSION
        assert frame.data == {"foo": 0}


def test_frame_length_without_data(frames: tuple[Request, Response]) -> None:
    """Test a frame length without any data."""
    for frame in frames:
        assert frame.length == struct_header.size + 3
        assert len(frame) == struct_header.size + 3


def test_get_header(frames: tuple[Request, Response]) -> None:
    """Test getting a frame header as bytes."""
    for frame in frames:
        assert frame.header == b"\x68\x0a\x00\x00\x56\x30\x05"


def test_base_class_with_message() -> None:
    """Test base request class with a message."""
    frame = RequestFrame(message=bytearray.fromhex("B00B"))
    assert frame.message == b"\xb0\x0b"


def test_to_bytes() -> None:
    """Test conversion to bytes."""
    frame = RequestFrame(message=bytearray(b"\xb0\x0b"))
    assert frame.bytes == b"\x68\x0c\x00\x00\x56\x30\x05\x40\xb0\x0b\xfc\x16"


def test_to_hex() -> None:
    """Test conversion to hex."""
    frame = RequestFrame(message=bytearray(b"\xb0\x0b"))
    hex_data = "680c000056300540b00bfc16"
    assert frame.hex() == hex_data


def test_equality() -> None:
    """Test frame objects equality."""
    assert ProgramVersionResponse() == ProgramVersionResponse()
    assert ProgramVersionResponse().__eq__("melon") is NotImplemented


def test_data_setter():
    """Test frame data setter."""
    frame = ProgramVersionResponse(data={ATTR_VERSION: VersionInfo(software="1.0.0")})
    assert frame.message.hex() == "ffff057a0000000001000000000056"
    frame.data = {ATTR_VERSION: VersionInfo(software="1.0.0", struct_version=6)}
    assert frame.message.hex() == "ffff067a0000000001000000000056"


def test_message_setter():
    """Test frame message setter."""
    frame = ProgramVersionResponse(
        message=bytearray.fromhex("ffff057a00000000000004000f0056")
    )
    assert frame.data[ATTR_VERSION].struct_version == 5
    frame.message = bytearray.fromhex("ffff067a00000000000004000f0056")
    assert frame.data[ATTR_VERSION].struct_version == 6


def test_request_repr(request_frame: Request) -> None:
    """Test a request class serialiazible representation."""
    assert repr(request_frame) == (
        "RequestFrame("
        "recipient=<DeviceType.ALL: 0>, "
        "sender=<DeviceType.ECONET: 86>, "
        "econet_type=48, "
        "econet_version=5, "
        "message=bytearray(b''), "
        "data={'foo': 0})"
    )


def test_response_repr(response_frame: Response) -> None:
    """Test a response class serialiazible representation."""
    assert repr(response_frame) == (
        "ResponseFrame("
        "recipient=<DeviceType.ALL: 0>, "
        "sender=<DeviceType.ECONET: 86>, "
        "econet_type=48, "
        "econet_version=5, "
        "message=bytearray(b''), "
        "data={'foo': 0})"
    )



"""Contains tests for the schedule helper classes."""

import asyncio
from unittest.mock import Mock, patch

import pytest

from pyplumio.const import (
    ATTR_PARAMETER,
    ATTR_SCHEDULE,
    ATTR_SWITCH,
    ATTR_TYPE,
    STATE_ON,
    FrameType,
)
from pyplumio.devices import Device, DeviceType
from pyplumio.helpers.schedule import Schedule, ScheduleDay
from pyplumio.structures.schedules import (
    ATTR_SCHEDULE_PARAMETER,
    ATTR_SCHEDULE_SWITCH,
    ATTR_SCHEDULES,
)


@pytest.fixture(name="schedule_day")
def fixture_schedule_day() -> ScheduleDay:
    """Return a schedule day object."""
    return ScheduleDay([False for _ in range(48)])


@pytest.fixture(name="schedule")
def fixture_schedule(schedule_day: ScheduleDay) -> Schedule:
    """Return a schedule object."""
    return Schedule(
        name="test",
        device=Mock(spec=Device),
        monday=ScheduleDay([True for _ in range(48)]),
        tuesday=schedule_day,
        wednesday=schedule_day,
        thursday=schedule_day,
        friday=schedule_day,
        saturday=schedule_day,
        sunday=schedule_day,
    )


def test_schedule_day(schedule_day: ScheduleDay) -> None:
    """Test a schedule day."""
    schedule_day.set_state(STATE_ON, "00:00", "01:00")
    assert schedule_day.intervals[0]  # 00:00
    assert schedule_day.intervals[1]  # 00:30
    assert schedule_day.intervals[2]  # 01:00
    assert not schedule_day.intervals[3]  # 01:30
    assert len(schedule_day) == 48


def test_schedule_day_with_incorrect_interval(schedule_day: ScheduleDay) -> None:
    """Test a schedule day with an incorrect interval."""
    for start, end in (("01:00", "00:30"), ("00:foo", "bar")):
        with pytest.raises(ValueError):
            schedule_day.set_state(STATE_ON, start, end)


def test_schedule_day_with_incorrect_state(schedule_day: ScheduleDay) -> None:
    """Test a schedule day with an incorrect state."""
    with pytest.raises(ValueError):
        schedule_day.set_state("invalid_state", "00:00", "01:00")  # type: ignore [arg-type]


def test_setting_whole_day_schedule(schedule_day: ScheduleDay) -> None:
    """Test setting schedule for a whole day."""
    schedule_day.set_on()
    assert schedule_day.intervals == [True for _ in range(48)]
    schedule_day.set_off()
    assert schedule_day.intervals == [False for _ in range(48)]


def test_setting_schedule_with_sequence(schedule_day: ScheduleDay) -> None:
    """Test setting a schedule via sequence methods."""
    schedule_day.set_on("00:30", "01:00")
    schedule_day_iter = iter(schedule_day)
    assert not next(schedule_day_iter)
    assert next(schedule_day_iter)
    assert schedule_day[1]
    schedule_day[1] = False
    del schedule_day[0]
    assert not schedule_day[0]
    schedule_day.append(True)
    assert schedule_day[-1]


def test_schedule_day_repr(schedule_day: ScheduleDay) -> None:
    """Test serializable representation of a schedule day."""
    assert repr(schedule_day) == f"ScheduleDay({[False for _ in range(48)]})"


def test_schedule(schedule: Schedule) -> None:
    """Test a schedule."""
    schedule_iter = iter(schedule)
    assert not next(schedule_iter)[0]
    assert next(schedule_iter)[0]


@patch("pyplumio.helpers.schedule.Request.create")
async def test_schedule_commit(mock_request_create, schedule: Schedule) -> None:
    """Test committing a schedule."""
    schedule.device = Mock(spec=Device)
    schedule.device.address = DeviceType.ECOMAX
    schedule.device.data = {
        f"test_{ATTR_SCHEDULE_SWITCH}": 1,
        f"test_{ATTR_SCHEDULE_PARAMETER}": 2,
        ATTR_SCHEDULES: {"test": schedule},
    }
    schedule.device.queue = Mock(spec=asyncio.Queue)

    await schedule.commit()
    mock_request_create.assert_awaited_once_with(
        FrameType.REQUEST_SET_SCHEDULE,
        recipient=DeviceType.ECOMAX,
        data={
            ATTR_TYPE: "test",
            ATTR_SWITCH: 1,
            ATTR_PARAMETER: 2,
            ATTR_SCHEDULE: schedule,
        },
    )



"""Contains tests for the task manager helper class."""

import asyncio
from unittest.mock import AsyncMock, Mock, patch

import pytest

from pyplumio.helpers.task_manager import TaskManager


@pytest.fixture(name="task_manager")
async def fixture_task_manager() -> TaskManager:
    """Return a task manager object."""
    task_manager = TaskManager()
    with patch("asyncio.create_task"):
        task_manager.create_task(Mock())

    return task_manager


def test_create_task(task_manager: TaskManager) -> None:
    """Test creating a task."""
    mock_coro = Mock()
    mock_task = Mock(spec=asyncio.Task)
    with patch("asyncio.create_task", return_value=mock_task) as create_task_mock:
        task_manager.create_task(mock_coro, name="test_task")

    mock_task.add_done_callback.assert_called_once()
    create_task_mock.assert_called_once_with(mock_coro, name="test_task")


def test_cancel_task(task_manager: TaskManager) -> None:
    """Test canceling a task."""
    mock_coro = Mock()
    mock_task = Mock(spec=asyncio.Task)
    with patch("asyncio.create_task", return_value=mock_task):
        task_manager.create_task(mock_coro)

    task_manager.cancel_tasks()
    mock_task.cancel.assert_called_once()


async def test_wait_until_done(task_manager: TaskManager) -> None:
    """Test waiting until all tasks are done."""
    with patch("asyncio.gather", new_callable=AsyncMock) as mock_gather:
        await task_manager.wait_until_done()

    mock_gather.assert_awaited_once_with(*task_manager.tasks, return_exceptions=True)



"""Contains device classes."""

from __future__ import annotations

from abc import ABC
import asyncio
from collections.abc import Iterable
from functools import cache
from typing import Any, ClassVar

from pyplumio.const import ATTR_FRAME_ERRORS, ATTR_LOADED, DeviceType, FrameType
from pyplumio.exceptions import UnknownDeviceError
from pyplumio.frames import DataFrameDescription, Frame, Request
from pyplumio.helpers.event_manager import EventManager
from pyplumio.helpers.factory import create_instance
from pyplumio.helpers.parameter import SET_RETRIES, Parameter, ParameterValueType
from pyplumio.structures.network_info import NetworkInfo
from pyplumio.utils import to_camelcase


@cache
def is_known_device_type(device_type: int) -> bool:
    """Check if device type is known."""
    try:
        DeviceType(device_type)
        return True
    except ValueError:
        return False


@cache
def get_device_handler(device_type: int) -> str:
    """Return module name and class name for a given device type."""
    if not is_known_device_type(device_type):
        raise UnknownDeviceError(f"Unknown device type ({device_type})")

    type_name = to_camelcase(
        DeviceType(device_type).name,
        overrides={"ecomax": "EcoMAX", "ecoster": "EcoSTER"},
    )
    return f"devices.{type_name.lower()}.{type_name}"


class Device(ABC, EventManager):
    """Represents a device."""

    queue: asyncio.Queue[Frame]

    def __init__(self, queue: asyncio.Queue[Frame]):
        """Initialize a new device."""
        super().__init__()
        self.queue = queue

    async def set(
        self,
        name: str,
        value: ParameterValueType,
        timeout: float | None = None,
        retries: int = SET_RETRIES,
    ) -> bool:
        """Set a parameter value.

        :param name: Name of the parameter
        :type name: str
        :param value: New value for the parameter
        :type value: int | float | bool | Literal["off", "on"]
        :param timeout: Wait this amount of seconds for confirmation,
            defaults to `None`
        :type timeout: float, optional
        :param retries: Try setting parameter for this amount of
            times, defaults to 5
        :type retries: int, optional
        :return: `True` if parameter was successfully set, `False`
            otherwise.
        :rtype: bool
        :raise asyncio.TimeoutError: when waiting past specified timeout
        :raise ValueError: when a new value is outside of allowed range
        :raise TypeError: when found data is not valid parameter
        """
        parameter = await self.get(name, timeout=timeout)
        if not isinstance(parameter, Parameter):
            raise TypeError(f"{name} is not valid parameter")

        return await parameter.set(value, retries=retries)

    def set_nowait(
        self,
        name: str,
        value: ParameterValueType,
        timeout: float | None = None,
        retries: int = SET_RETRIES,
    ) -> None:
        """Set a parameter value without waiting for the result.

        :param name: Name of the parameter
        :type name: str
        :param value: New value for the parameter
        :type value: int | float | bool | Literal["off", "on"]
        :param timeout: Wait this amount of seconds for confirmation.
            As this method operates in the background without waiting,
            this value is used to determine failure when
            retrying and doesn't block, defaults to `None`
        :type timeout: float, optional
        :param retries: Try setting parameter for this amount of
            times, defaults to 5
        :type retries: int, optional
        :return: `True` if parameter was successfully set, `False`
            otherwise.
        :rtype: bool
        """
        self.create_task(self.set(name, value, timeout, retries))

    async def shutdown(self) -> None:
        """Cancel device tasks."""
        self.cancel_tasks()
        await self.wait_until_done()


class AddressableDevice(Device, ABC):
    """Represents an addressable device."""

    address: ClassVar[int]
    _network: NetworkInfo
    _setup_frames: Iterable[DataFrameDescription]

    def __init__(self, queue: asyncio.Queue[Frame], network: NetworkInfo):
        """Initialize a new addressable device."""
        super().__init__(queue)
        self._network = network

    def __int__(self) -> int:
        """Return the device address."""
        return int(self.address)

    def handle_frame(self, frame: Frame) -> None:
        """Handle frame received from the device."""
        frame.sender_device = self
        if frame.data is not None:
            for name, value in frame.data.items():
                self.dispatch_nowait(name, value)

    async def async_setup(self) -> bool:
        """Set up addressable device."""
        results = await asyncio.gather(
            *(
                self.request(description.provides, description.frame_type)
                for description in self._setup_frames
            ),
            return_exceptions=True,
        )

        errors = [
            result.args[1] for result in results if isinstance(result, BaseException)
        ]

        await asyncio.gather(
            self.dispatch(ATTR_FRAME_ERRORS, errors), self.dispatch(ATTR_LOADED, True)
        )
        return True

    async def request(
        self, name: str, frame_type: FrameType, retries: int = 3, timeout: float = 3.0
    ) -> Any:
        """Send request and wait for a value to become available.

        If value is not available before timeout, retry request.
        """
        request = await Request.create(frame_type, recipient=self.address)
        while retries > 0:
            try:
                self.queue.put_nowait(request)
                return await self.get(name, timeout=timeout)
            except asyncio.TimeoutError:
                retries -= 1

        raise ValueError(f'could not request "{name}"', frame_type)

    @classmethod
    async def create(cls, device_type: int, **kwargs: Any) -> AddressableDevice:
        """Create a device handler object."""
        return await create_instance(get_device_handler(device_type), cls=cls, **kwargs)


class SubDevice(Device, ABC):
    """Represents a sub-device."""

    parent: AddressableDevice
    index: int

    def __init__(
        self, queue: asyncio.Queue[Frame], parent: AddressableDevice, index: int = 0
    ):
        """Initialize a new sub-device."""
        super().__init__(queue)
        self.parent = parent
        self.index = index



"""Contains exceptions."""

from __future__ import annotations


class PyPlumIOError(Exception):
    """Base PyPlumIO error class."""


class ConnectionFailedError(PyPlumIOError):
    """Raised on connection failure."""


class ProtocolError(PyPlumIOError):
    """Base class for protocol-related errors."""


class ReadError(ProtocolError):
    """Raised on read error."""


class ChecksumError(ProtocolError):
    """Raised on incorrect frame checksum."""


class UnknownDeviceError(ProtocolError):
    """Raised on unknown device."""


class UnknownFrameError(ProtocolError):
    """Raised on unknown frame type."""


class FrameDataError(ProtocolError):
    """Raised on incorrect frame data."""



"""Contains a factory helper."""

from __future__ import annotations

import asyncio
import importlib
import logging
from types import ModuleType
from typing import Any, TypeVar

_LOGGER = logging.getLogger(__name__)

T = TypeVar("T")


async def _load_module(module_name: str) -> ModuleType:
    """Load a module by name."""
    return await asyncio.get_running_loop().run_in_executor(
        None, importlib.import_module, f".{module_name}", "pyplumio"
    )


async def create_instance(class_path: str, cls: type[T], **kwargs: Any) -> T:
    """Return a class instance from the class path."""
    module_name, class_name = class_path.rsplit(".", 1)
    try:
        module = await _load_module(module_name)
        instance = getattr(module, class_name)(**kwargs)
        if not isinstance(instance, cls):
            raise TypeError(f"class '{class_name}' should be derived from {cls}")

        return instance
    except Exception:
        _LOGGER.error("Failed to load module (%s)", class_path)
        raise



"""Contains schedule decoder."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from functools import reduce
from itertools import chain
from typing import Any, Final

from dataslots import dataslots

from pyplumio.const import (
    ATTR_PARAMETER,
    ATTR_SCHEDULE,
    ATTR_SWITCH,
    ATTR_TYPE,
    FrameType,
)
from pyplumio.devices import AddressableDevice, Device
from pyplumio.exceptions import FrameDataError
from pyplumio.frames import Request
from pyplumio.helpers.parameter import (
    Number,
    NumberDescription,
    Parameter,
    ParameterDescription,
    ParameterValues,
    Switch,
    SwitchDescription,
    unpack_parameter,
)
from pyplumio.structures import Structure
from pyplumio.utils import ensure_dict

ATTR_SCHEDULES: Final = "schedules"
ATTR_SCHEDULE_PARAMETERS: Final = "schedule_parameters"
ATTR_SCHEDULE_SWITCH: Final = "schedule_switch"
ATTR_SCHEDULE_PARAMETER: Final = "schedule_parameter"

SCHEDULE_SIZE: Final = 42  # 6 bytes per day, 7 days total

SCHEDULES: tuple[str, ...] = (
    "heating",
    "water_heater",
    "circulation_pump",
    "boiler_work",
    "boiler_clean",
    "hear_exchanger_clean",
    "mixer_1",
    "mixer_2",
    "mixer_3",
    "mixer_4",
    "mixer_5",
    "mixer_6",
    "mixer_7",
    "mixer_8",
    "mixer_9",
    "mixer_10",
    "thermostat_1",
    "thermostat_2",
    "thermostat_3",
    "circuit_1",
    "circuit_2",
    "circuit_3",
    "circuit_4",
    "circuit_5",
    "circuit_6",
    "circuit_7",
    "panel_1",
    "panel_2",
    "panel_3",
    "panel_4",
    "panel_5",
    "panel_6",
    "panel_7",
    "main_heater_solar",
    "heating_circulation",
    "internal_thermostat",
    "heater",
    "water_heater_2",
    "intake",
    "intake_summer",
)


@dataclass
class ScheduleParameterDescription(ParameterDescription):
    """Represent a schedule parameter description."""

    __slots__ = ()


class ScheduleParameter(Parameter):
    """Represents a schedule parameter."""

    __slots__ = ()

    device: AddressableDevice
    description: ScheduleParameterDescription

    async def create_request(self) -> Request:
        """Create a request to change the parameter."""
        schedule_name, _ = self.description.name.split("_schedule_", 1)
        return await Request.create(
            FrameType.REQUEST_SET_SCHEDULE,
            recipient=self.device.address,
            data=collect_schedule_data(schedule_name, self.device),
        )


@dataslots
@dataclass
class ScheduleNumberDescription(ScheduleParameterDescription, NumberDescription):
    """Represents a schedule number description."""


class ScheduleNumber(ScheduleParameter, Number):
    """Represents a schedule number."""

    __slots__ = ()

    description: ScheduleNumberDescription


@dataslots
@dataclass
class ScheduleSwitchDescription(ScheduleParameterDescription, SwitchDescription):
    """Represents a schedule switch description."""


class ScheduleSwitch(ScheduleParameter, Switch):
    """Represents a schedule switch."""

    __slots__ = ()

    description: ScheduleSwitchDescription


SCHEDULE_PARAMETERS: list[ScheduleParameterDescription] = list(
    chain.from_iterable(
        [
            [
                ScheduleSwitchDescription(name=f"{name}_{ATTR_SCHEDULE_SWITCH}"),
                ScheduleNumberDescription(name=f"{name}_{ATTR_SCHEDULE_PARAMETER}"),
            ]
            for name in SCHEDULES
        ]
    )
)


def collect_schedule_data(name: str, device: Device) -> dict[str, Any]:
    """Return a schedule data collected from the device."""
    return {
        ATTR_TYPE: name,
        ATTR_SWITCH: device.data[f"{name}_{ATTR_SCHEDULE_SWITCH}"],
        ATTR_PARAMETER: device.data[f"{name}_{ATTR_SCHEDULE_PARAMETER}"],
        ATTR_SCHEDULE: device.data[ATTR_SCHEDULES][name],
    }


def _split_byte(byte: int) -> list[bool]:
    """Split single byte into an eight bits."""
    return [bool(byte & (1 << bit)) for bit in reversed(range(8))]


def _join_bits(bits: Sequence[int | bool]) -> int:
    """Join eight bits into a single byte."""
    return reduce(lambda x, y: (x << 1) | y, bits)


class SchedulesStructure(Structure):
    """Represents a schedule data structure."""

    __slots__ = ("_offset",)

    _offset: int

    def encode(self, data: dict[str, Any]) -> bytearray:
        """Encode data to the bytearray message."""
        message = bytearray([1])
        try:
            message.append(SCHEDULES.index(data[ATTR_TYPE]))
            message.append(int(data[ATTR_SWITCH]))
            message.append(int(data[ATTR_PARAMETER]))
            schedule = data[ATTR_SCHEDULE]
        except (KeyError, ValueError) as e:
            raise FrameDataError from e

        return message + bytearray(
            chain.from_iterable(
                [_join_bits(day[i : i + 8]) for i in range(0, len(day), 8)]
                for day in list(schedule)
            )
        )

    def _unpack_schedule(self, message: bytearray) -> list[list[bool]]:
        """Unpack a schedule."""
        schedule: list[bool] = []
        last_offset = self._offset + SCHEDULE_SIZE
        while self._offset < last_offset:
            schedule += _split_byte(message[self._offset])
            self._offset += 1

        # Split schedule. Each day consists of 48 half-hour intervals.
        return [schedule[i : i + 48] for i in range(0, len(schedule), 48)]

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        try:
            offset += 1
            start = message[offset]
            offset += 1
            end = message[offset]
        except IndexError:
            return ensure_dict(data, {ATTR_SCHEDULES: []}), offset

        self._offset = offset + 1
        schedules: list[tuple[int, list[list[bool]]]] = []
        parameters: list[tuple[int, ParameterValues]] = []

        for _ in range(start, start + end):
            index = message[self._offset]
            switch = ParameterValues(
                value=message[self._offset + 1], min_value=0, max_value=1
            )
            parameter = unpack_parameter(message, self._offset + 2)
            self._offset += 5
            schedules.append((index, self._unpack_schedule(message)))
            parameters.append((index * 2, switch))
            if parameter is not None:
                parameters.append((index * 2 + 1, parameter))

        return (
            ensure_dict(
                data, {ATTR_SCHEDULES: schedules, ATTR_SCHEDULE_PARAMETERS: parameters}
            ),
            self._offset,
        )



"""Contains a fan power structure decoder."""

from __future__ import annotations

import math
from typing import Any, Final

from pyplumio.helpers.data_types import Float
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_FAN_POWER: Final = "fan_power"


class FanPowerStructure(StructureDecoder):
    """Represents a fan power sensor data structure."""

    __slots__ = ()

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        fan_power = Float.from_bytes(message, offset)
        offset += fan_power.size

        if math.isnan(fan_power.value):
            return ensure_dict(data), offset

        return ensure_dict(data, {ATTR_FAN_POWER: fan_power.value}), offset



"""Contains an ecoSTER class."""

from __future__ import annotations

from typing import ClassVar

from pyplumio.const import DeviceType
from pyplumio.devices import AddressableDevice


class EcoSTER(AddressableDevice):
    """Represents an ecoSTER thermostat."""

    address: ClassVar[int] = DeviceType.ECOSTER



"""Contains a device parameter class."""

from __future__ import annotations

from abc import ABC, abstractmethod
import asyncio
from dataclasses import dataclass
import logging
from typing import TYPE_CHECKING, Any, Final, Literal, TypeVar, Union

from dataslots import dataslots
from typing_extensions import TypeAlias

from pyplumio.const import BYTE_UNDEFINED, STATE_OFF, STATE_ON, UnitOfMeasurement
from pyplumio.frames import Request

if TYPE_CHECKING:
    from pyplumio.devices import Device

_LOGGER = logging.getLogger(__name__)

SET_TIMEOUT: Final = 5
SET_RETRIES: Final = 5

ParameterValueType: TypeAlias = Union[int, float, bool, Literal["off", "on"]]
ParameterT = TypeVar("ParameterT", bound="Parameter")


def unpack_parameter(
    data: bytearray, offset: int = 0, size: int = 1
) -> ParameterValues | None:
    """Unpack a device parameter."""
    if not check_parameter(data[offset : offset + size * 3]):
        return None

    value = data[offset : offset + size]
    min_value = data[offset + size : offset + 2 * size]
    max_value = data[offset + 2 * size : offset + 3 * size]

    return ParameterValues(
        value=int.from_bytes(value, byteorder="little"),
        min_value=int.from_bytes(min_value, byteorder="little"),
        max_value=int.from_bytes(max_value, byteorder="little"),
    )


def check_parameter(data: bytearray) -> bool:
    """Check if parameter contains any bytes besides 0xFF."""
    return any(x for x in data if x != BYTE_UNDEFINED)


def _normalize_parameter_value(value: ParameterValueType) -> int:
    """Normalize a parameter value."""
    if value in (STATE_OFF, STATE_ON):
        return 1 if value == STATE_ON else 0

    return int(value)


@dataclass
class ParameterValues:
    """Represents a parameter values."""

    __slots__ = ("value", "min_value", "max_value")

    value: int
    min_value: int
    max_value: int


@dataslots
@dataclass
class ParameterDescription:
    """Represents a parameter description."""

    name: str


class Parameter(ABC):
    """Represents a base parameter."""

    __slots__ = ("device", "description", "_pending_update", "_index", "_values")

    device: Device
    description: ParameterDescription
    _pending_update: bool
    _index: int
    _values: ParameterValues

    def __init__(
        self,
        device: Device,
        description: ParameterDescription,
        values: ParameterValues | None = None,
        index: int = 0,
    ):
        """Initialize a new parameter."""
        self.device = device
        self.description = description
        self._pending_update = False
        self._index = index
        self._values = values if values else ParameterValues(0, 0, 0)

    def __repr__(self) -> str:
        """Return a serializable string representation."""
        return (
            f"{self.__class__.__name__}("
            f"device={self.device.__class__.__name__}, "
            f"description={self.description}, "
            f"values={self.values}, "
            f"index={self._index})"
        )

    def _call_relational_method(self, method_to_call: str, other: Any) -> Any:
        """Call a specified relational method."""
        if isinstance(other, Parameter):
            other = other.values

        if isinstance(other, ParameterValues):
            handler = getattr(self.values, method_to_call)
            return handler(other)

        if isinstance(other, (int, float, bool)) or other in (STATE_OFF, STATE_ON):
            handler = getattr(self.values.value, method_to_call)
            return handler(_normalize_parameter_value(other))
        else:
            return NotImplemented

    def __int__(self) -> int:
        """Return an integer representation of parameter's value."""
        return self.values.value

    def __add__(self, other: Any) -> Any:
        """Return result of addition."""
        return self._call_relational_method("__add__", other)

    def __sub__(self, other: Any) -> Any:
        """Return result of the subtraction."""
        return self._call_relational_method("__sub__", other)

    def __truediv__(self, other: Any) -> Any:
        """Return result of true division."""
        return self._call_relational_method("__truediv__", other)

    def __floordiv__(self, other: Any) -> Any:
        """Return result of floored division."""
        return self._call_relational_method("__floordiv__", other)

    def __mul__(self, other: Any) -> Any:
        """Return result of the multiplication."""
        return self._call_relational_method("__mul__", other)

    def __eq__(self, other: Any) -> Any:
        """Compare if parameter value is equal to other."""
        return self._call_relational_method("__eq__", other)

    def __ge__(self, other: Any) -> Any:
        """Compare if parameter value is greater or equal to other."""
        return self._call_relational_method("__ge__", other)

    def __gt__(self, other: Any) -> Any:
        """Compare if parameter value is greater than other."""
        return self._call_relational_method("__gt__", other)

    def __le__(self, other: Any) -> Any:
        """Compare if parameter value is less or equal to other."""
        return self._call_relational_method("__le__", other)

    def __lt__(self, other: Any) -> Any:
        """Compare if parameter value is less that other."""
        return self._call_relational_method("__lt__", other)

    async def set(self, value: Any, retries: int = SET_RETRIES) -> bool:
        """Set a parameter value."""
        if (value := _normalize_parameter_value(value)) == self.values.value:
            return True

        if value < self.values.min_value or value > self.values.max_value:
            raise ValueError(
                f"Value must be between '{self.min_value}' and '{self.max_value}'"
            )

        self._values.value = value
        self._pending_update = True
        while self.pending_update:
            if retries <= 0:
                _LOGGER.error(
                    "Timed out while trying to set '%s' parameter",
                    self.description.name,
                )
                return False

            await self.device.queue.put(await self.create_request())
            await asyncio.sleep(SET_TIMEOUT)
            retries -= 1

        return True

    def update(self, values: ParameterValues) -> None:
        """Update the parameter values."""
        self._values = values
        self._pending_update = False

    @property
    def pending_update(self) -> bool:
        """Check if parameter is pending update on the device."""
        return self._pending_update

    @property
    def values(self) -> ParameterValues:
        """Return the parameter values."""
        return self._values

    @classmethod
    def create_or_update(
        cls: type[ParameterT],
        device: Device,
        description: ParameterDescription,
        values: ParameterValues,
        **kwargs: Any,
    ) -> ParameterT:
        """Create new parameter or update parameter values."""
        parameter: ParameterT | None = device.get_nowait(description.name, None)
        if parameter and isinstance(parameter, cls):
            parameter.update(values)
        else:
            parameter = cls(
                device=device, description=description, values=values, **kwargs
            )

        return parameter

    @property
    @abstractmethod
    def value(self) -> Any:
        """Return the value."""

    @property
    @abstractmethod
    def min_value(self) -> Any:
        """Return the minimum allowed value."""

    @property
    @abstractmethod
    def max_value(self) -> Any:
        """Return the maximum allowed value."""

    @abstractmethod
    async def create_request(self) -> Request:
        """Create a request to change the parameter."""


@dataslots
@dataclass
class NumberDescription(ParameterDescription):
    """Represents a parameter description."""

    unit_of_measurement: UnitOfMeasurement | Literal["%"] | None = None


class Number(Parameter):
    """Represents a number."""

    __slots__ = ()

    description: NumberDescription

    async def set(self, value: int | float, retries: int = SET_RETRIES) -> bool:
        """Set a parameter value."""
        return await super().set(value, retries)

    def set_nowait(self, value: int | float, retries: int = SET_RETRIES) -> None:
        """Set a parameter value without waiting."""
        self.device.create_task(self.set(value, retries))

    async def create_request(self) -> Request:
        """Create a request to change the number."""
        return Request()

    @property
    def value(self) -> int | float:
        """Return the value."""
        return self.values.value

    @property
    def min_value(self) -> int | float:
        """Return the minimum allowed value."""
        return self.values.min_value

    @property
    def max_value(self) -> int | float:
        """Return the maximum allowed value."""
        return self.values.max_value

    @property
    def unit_of_measurement(self) -> UnitOfMeasurement | Literal["%"] | None:
        """Return the unit of measurement."""
        return self.description.unit_of_measurement


@dataslots
@dataclass
class SwitchDescription(ParameterDescription):
    """Represents a switch description."""


class Switch(Parameter):
    """Represents a switch."""

    __slots__ = ()

    description: SwitchDescription

    async def set(
        self, value: bool | Literal["off", "on"], retries: int = SET_RETRIES
    ) -> bool:
        """Set a parameter value."""
        return await super().set(value, retries)

    def set_nowait(
        self, value: bool | Literal["off", "on"], retries: int = SET_RETRIES
    ) -> None:
        """Set a switch value without waiting."""
        self.device.create_task(self.set(value, retries))

    async def turn_on(self) -> bool:
        """Set a switch value to 'on'.

        :return: `True` if parameter was successfully turned on, `False`
            otherwise.
        :rtype: bool
        """
        return await self.set(STATE_ON)

    async def turn_off(self) -> bool:
        """Set a switch value to 'off'.

        :return: `True` if parameter was successfully turned off, `False`
            otherwise.
        :rtype: bool
        """
        return await self.set(STATE_OFF)

    def turn_on_nowait(self) -> None:
        """Set a switch value to 'on' without waiting."""
        self.set_nowait(STATE_ON)

    def turn_off_nowait(self) -> None:
        """Set a switch value to 'off' without waiting."""
        self.set_nowait(STATE_OFF)

    async def create_request(self) -> Request:
        """Create a request to change the switch."""
        return Request()

    @property
    def value(self) -> Literal["off", "on"]:
        """Return the value."""
        return STATE_ON if self.values.value == 1 else STATE_OFF

    @property
    def min_value(self) -> Literal["off"]:
        """Return the minimum allowed value."""
        return STATE_OFF

    @property
    def max_value(self) -> Literal["on"]:
        """Return the maximum allowed value."""
        return STATE_ON



"""Contains data schema structure decoder."""

from __future__ import annotations

from typing import Any, Final

from pyplumio.helpers.data_types import DATA_TYPES, DataType, UnsignedShort
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_REGDATA_SCHEMA: Final = "regdata_schema"


class RegulatorDataSchemaStructure(StructureDecoder):
    """Represents a regulator data schema structure."""

    __slots__ = ("_offset",)

    _offset: int

    def _unpack_block(self, message: bytearray) -> tuple[int, DataType]:
        """Unpack a block."""
        param_type = message[self._offset]
        self._offset += 1
        param_id = UnsignedShort.from_bytes(message, self._offset)
        self._offset += param_id.size
        return param_id.value, DATA_TYPES[param_type]()

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        blocks = UnsignedShort.from_bytes(message, offset)
        self._offset = offset + blocks.size
        if blocks.value == 0:
            return ensure_dict(data), self._offset

        return (
            ensure_dict(
                data,
                {
                    ATTR_REGDATA_SCHEMA: [
                        self._unpack_block(message) for _ in range(blocks.value)
                    ]
                },
            ),
            self._offset,
        )



"""Contains a program version decoder."""

from __future__ import annotations

from dataclasses import dataclass
import struct
from typing import Any, Final

from pyplumio._version import __version_tuple__
from pyplumio.structures import Structure
from pyplumio.utils import ensure_dict

ATTR_VERSION: Final = "version"

VERSION_INFO_SIZE: Final = 15

SOFTWARE_VERSION: Final = ".".join(str(x) for x in __version_tuple__[0:3])

struct_program_version = struct.Struct("<2sB2s3s3HB")


@dataclass
class VersionInfo:
    """Represents a version info provided in program version response."""

    software: str = SOFTWARE_VERSION
    struct_tag: bytes = b"\xff\xff"
    struct_version: int = 5
    device_id: bytes = b"\x7a\x00"
    processor_signature: bytes = b"\x00\x00\x00"


class ProgramVersionStructure(Structure):
    """Represents a program version data structure."""

    __slots__ = ()

    def encode(self, data: dict[str, Any]) -> bytearray:
        """Encode data to the bytearray message."""
        message = bytearray(struct_program_version.size)
        version_info: VersionInfo = data.get(ATTR_VERSION, VersionInfo())
        struct_program_version.pack_into(
            message,
            0,
            version_info.struct_tag,
            version_info.struct_version,
            version_info.device_id,
            version_info.processor_signature,
            *map(int, version_info.software.split(".", 2)),
            self.frame.sender,
        )
        return message

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        (
            struct_tag,
            struct_version,
            device_id,
            processor_signature,
            software1,
            software2,
            software3,
            _,  # recipient
        ) = struct_program_version.unpack_from(message)

        return (
            ensure_dict(
                data,
                {
                    ATTR_VERSION: VersionInfo(
                        software=".".join(map(str, [software1, software2, software3])),
                        struct_tag=struct_tag,
                        struct_version=struct_version,
                        device_id=device_id,
                        processor_signature=processor_signature,
                    )
                },
            ),
            offset + VERSION_INFO_SIZE,
        )



"""Contains tests for the frame reader and writer classes."""

import asyncio
from collections.abc import Generator
from unittest.mock import AsyncMock, patch

import pytest

from pyplumio.const import DeviceType, FrameType
from pyplumio.exceptions import ChecksumError, ReadError, UnknownDeviceError
from pyplumio.frames import ECONET_TYPE, ECONET_VERSION
from pyplumio.frames.requests import EcomaxParametersRequest, ProgramVersionRequest
from pyplumio.stream import FrameReader, FrameWriter


@pytest.fixture(name="frame_reader")
def fixture_frame_reader() -> Generator[FrameReader, None, None]:
    """Return instance of frame reader."""
    with patch("asyncio.StreamReader") as mock_stream_reader:
        yield FrameReader(mock_stream_reader)


@patch("asyncio.StreamWriter", autospec=True)
async def test_frame_writer(mock_stream_writer) -> None:
    """Test frame writer."""
    frame = ProgramVersionRequest()
    writer = FrameWriter(mock_stream_writer)
    await writer.write(frame)
    mock_stream_writer.write.assert_called_once_with(frame.bytes)
    mock_stream_writer.drain.assert_awaited_once()
    await writer.close()
    mock_stream_writer.close.assert_called_once()
    mock_stream_writer.wait_closed.assert_awaited_once()


@pytest.mark.parametrize(
    ("method", "exception"), [("close", OSError), ("wait_closed", asyncio.TimeoutError)]
)
@patch("asyncio.StreamWriter", autospec=True)
async def test_frame_writer_with_close_error(
    mock_stream_writer, method, exception, caplog
) -> None:
    """Test frame writer with error on writer close."""
    writer = FrameWriter(mock_stream_writer)
    getattr(mock_stream_writer, method).side_effect = exception
    await writer.close()
    assert "Unexpected error while closing the writer" in caplog.text


@patch(
    "asyncio.StreamReader.read",
    side_effect=(b"\x00", b"\x68", b"\x0c\x00\x00\x56\x30\x05"),
    new_callable=AsyncMock,
)
@patch(
    "asyncio.StreamReader.readexactly",
    return_value=b"\x31\xff\x00\xc9\x16",
    new_callable=AsyncMock,
)
async def test_frame_reader(
    mock_readexactly, mock_read, frame_reader: FrameReader
) -> None:
    """Test frame reader."""
    frame = await frame_reader.read()
    assert isinstance(frame, EcomaxParametersRequest)
    assert frame.frame_type == FrameType.REQUEST_ECOMAX_PARAMETERS
    assert frame.sender == DeviceType.ECONET
    assert frame.econet_type == ECONET_TYPE
    assert frame.recipient == DeviceType.ALL
    assert frame.message == b"\xff\x00"
    assert frame.econet_version == ECONET_VERSION
    assert mock_read.call_count == 3
    assert mock_readexactly.call_count == 1


@patch(
    "asyncio.StreamReader.read",
    return_value=False,
    new_callable=AsyncMock,
)
async def test_frame_reader_with_empty_buffer(
    mock_read, frame_reader: FrameReader
) -> None:
    """Test reader with empty read buffer."""
    with pytest.raises(OSError):
        await frame_reader.read()

    mock_read.assert_awaited_once()


@patch(
    "asyncio.StreamReader.read",
    side_effect=(b"\x68", b"\x03\x00\x00"),
    new_callable=AsyncMock,
)
async def test_frame_reader_with_short_header(
    mock_read, frame_reader: FrameReader
) -> None:
    """Test reader when data is less than header size."""
    with pytest.raises(ReadError):
        await frame_reader.read()

    assert mock_read.call_count == 2


@patch(
    "asyncio.StreamReader.read",
    side_effect=(b"\x68", b"\x03\x00\x00\x56\x30\x05"),
    new_callable=AsyncMock,
)
@patch(
    "asyncio.StreamReader.readexactly",
    return_value=b"\x31\xff\x00\xc9\x16",
    new_callable=AsyncMock,
)
async def test_frame_reader_with_incorrect_length(
    mock_readexactly, mock_read, frame_reader: FrameReader
) -> None:
    """Test reader on frame with incorrect length."""
    with pytest.raises(ReadError):
        await frame_reader.read()


@patch(
    "asyncio.StreamReader.read",
    side_effect=(b"\x68", b"\x0c\x00\x00\x56\x30\x05"),
    new_callable=AsyncMock,
)
@patch(
    "asyncio.StreamReader.readexactly",
    side_effect=asyncio.IncompleteReadError(bytearray(), None),
    new_callable=AsyncMock,
)
async def test_frame_reader_with_incomplete_read(
    mock_readexactly, mock_read, frame_reader: FrameReader
) -> None:
    """Test reader with incomplete data."""
    with pytest.raises(ReadError):
        await frame_reader.read()


@patch(
    "asyncio.StreamReader.read",
    side_effect=(b"\x68", b"\x0c\x00\x00\x56\x30\x05"),
    new_callable=AsyncMock,
)
@patch(
    "asyncio.StreamReader.readexactly",
    return_value=b"\x31\xfe\x00\xc9\x16",
    new_callable=AsyncMock,
)
async def test_frame_reader_with_incorrect_bcc(
    mock_readexactly, mock_read, frame_reader: FrameReader
) -> None:
    """Test reader on frame with incorrect checksum."""
    with pytest.raises(ChecksumError):
        await frame_reader.read()


@patch(
    "asyncio.StreamReader.read",
    side_effect=(b"\x68", b"\x0c\x00\x10\x56\x30\x05"),
    new_callable=AsyncMock,
)
@patch(
    "asyncio.StreamReader.readexactly",
    return_value=b"\x31\xff\x00\xc9\x16",
    new_callable=AsyncMock,
)
async def test_frame_reader_with_unknown_recipient(
    mock_readexactly, mock_read, frame_reader: FrameReader
) -> None:
    """Test reader on frame with unknown recipient address."""
    result = await frame_reader.read()
    assert result is None


@patch(
    "asyncio.StreamReader.read",
    side_effect=(b"\x68", b"\x0c\x00\x00\x10\x30\x05"),
    new_callable=AsyncMock,
)
@patch(
    "asyncio.StreamReader.readexactly",
    return_value=b"\x31\xff\x00\xc9\x16",
    new_callable=AsyncMock,
)
async def test_frame_reader_with_unknown_sender(
    mock_readexactly, mock_read, frame_reader: FrameReader
) -> None:
    """Test reader on frame with unknown sender address."""
    with pytest.raises(UnknownDeviceError):
        await frame_reader.read()



"""Contains tests for the timeout decorator class."""

import asyncio
from unittest.mock import AsyncMock, Mock, patch

import pytest

from pyplumio.helpers.timeout import timeout


@patch(
    "asyncio.wait_for",
    new_callable=AsyncMock,
    side_effect=("test", asyncio.TimeoutError),
)
async def test_timeout(mock_wait_for) -> None:
    """Test a timeout decorator."""
    # Mock function to pass to the decorator.
    mock_func = Mock()
    mock_func.return_value = "test"

    # Call the decorator.
    decorator = timeout(10)
    wrapper = decorator(mock_func)
    result = await wrapper("test_arg", kwarg="test_kwarg")
    assert result == "test"
    mock_wait_for.assert_awaited_once_with("test", timeout=10)
    mock_func.assert_called_once_with("test_arg", kwarg="test_kwarg")

    # Check with raise_exception set to true.
    decorator = timeout(10)
    wrapper = decorator(mock_func)
    with pytest.raises(asyncio.TimeoutError):
        await wrapper("test_arg", kwarg="test_kwarg")



"""Contains a lambda sensor structure decoder."""

from __future__ import annotations

from contextlib import suppress
import math
from typing import Any, Final

from pyplumio.const import BYTE_UNDEFINED, LambdaState
from pyplumio.helpers.data_types import UnsignedShort
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_LAMBDA_STATE: Final = "lambda_state"
ATTR_LAMBDA_TARGET: Final = "lambda_target"
ATTR_LAMBDA_LEVEL: Final = "lambda_level"


class LambdaSensorStructure(StructureDecoder):
    """Represents a lambda sensor data structure."""

    __slots__ = ()

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        lambda_state = message[offset]
        offset += 1
        if lambda_state == BYTE_UNDEFINED:
            return ensure_dict(data), offset

        lambda_target = message[offset]
        offset += 1
        level = UnsignedShort.from_bytes(message, offset)
        offset += level.size
        with suppress(ValueError):
            lambda_state = LambdaState(lambda_state)

        return (
            ensure_dict(
                data,
                {
                    ATTR_LAMBDA_STATE: lambda_state,
                    ATTR_LAMBDA_TARGET: lambda_target,
                    ATTR_LAMBDA_LEVEL: (
                        None if math.isnan(level.value) else (level.value / 10)
                    ),
                },
            ),
            offset,
        )



"""Contains callback filters."""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from copy import copy
import math
import time
from typing import (
    Any,
    Final,
    Protocol,
    SupportsFloat,
    TypeVar,
    overload,
    runtime_checkable,
)

from pyplumio.helpers.event_manager import Callback
from pyplumio.helpers.parameter import Parameter

UNDEFINED: Final = "undefined"
TOLERANCE: Final = 0.1


@runtime_checkable
class SupportsSubtraction(Protocol):
    """Supports subtraction operation."""

    __slots__ = ()

    def __sub__(
        self: SupportsSubtraction, other: SupportsSubtraction
    ) -> SupportsSubtraction:
        """Subtract a value."""


@runtime_checkable
class SupportsComparison(Protocol):
    """Supports comparison."""

    __slots__ = ()

    def __eq__(self: SupportsComparison, other: SupportsComparison) -> bool:
        """Compare a value."""


Comparable = TypeVar("Comparable", Parameter, SupportsFloat, SupportsComparison)


@overload
def _significantly_changed(old: Parameter, new: Parameter) -> bool: ...


@overload
def _significantly_changed(old: SupportsFloat, new: SupportsFloat) -> bool: ...


@overload
def _significantly_changed(
    old: SupportsComparison, new: SupportsComparison
) -> bool: ...


def _significantly_changed(old: Comparable, new: Comparable) -> bool:
    """Check if value is significantly changed."""
    if isinstance(old, Parameter) and isinstance(new, Parameter):
        result = new.pending_update or old.values != new.values
    elif isinstance(old, SupportsFloat) and isinstance(new, SupportsFloat):
        result = not math.isclose(old, new, abs_tol=TOLERANCE)
    else:
        result = old != new

    return result


@overload
def _diffence_between(old: list, new: list) -> list: ...


@overload
def _diffence_between(
    old: SupportsSubtraction, new: SupportsSubtraction
) -> SupportsSubtraction: ...


def _diffence_between(
    old: SupportsSubtraction | list, new: SupportsSubtraction | list
) -> SupportsSubtraction | list | None:
    """Return a difference between values."""
    if isinstance(old, list) and isinstance(new, list):
        return [x for x in new if x not in old]
    elif isinstance(old, SupportsSubtraction) and isinstance(new, SupportsSubtraction):
        return new - old
    else:
        return None


class Filter(ABC):
    """Represents a filter."""

    __slots__ = ("_callback", "_value")

    _callback: Callback
    _value: Any

    def __init__(self, callback: Callback) -> None:
        """Initialize a new filter."""
        self._callback = callback
        self._value = UNDEFINED

    def __eq__(self, other: Any) -> bool:
        """Compare callbacks."""
        if isinstance(other, Filter):
            return self._callback == other._callback

        if callable(other):
            return bool(self._callback == other)

        return NotImplemented

    @abstractmethod
    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""


class _OnChange(Filter):
    """Represents a value changed filter.

    Calls a callback only when value is changed from the
    previous callback call.
    """

    __slots__ = ()

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if self._value == UNDEFINED or _significantly_changed(self._value, new_value):
            self._value = (
                copy(new_value) if isinstance(new_value, Parameter) else new_value
            )
            return await self._callback(new_value)


def on_change(callback: Callback) -> _OnChange:
    """Return a value changed filter.

    A callback function will only be called if value is changed from the
    previous call.

    :param callback: A callback function to be awaited on value change
    :type callback: Callable[[Any], Coroutine[Any, Any, Any]]
    :return: A instance of callable filter
    :rtype: _OnChange
    """
    return _OnChange(callback)


class _Debounce(Filter):
    """Represents a debounce filter.

    Calls a callback only when value is stabilized across multiple
    filter calls.
    """

    __slots__ = ("_calls", "_min_calls")

    _calls: int
    _min_calls: int

    def __init__(self, callback: Callback, min_calls: int) -> None:
        """Initialize a new debounce filter."""
        super().__init__(callback)
        self._calls = 0
        self._min_calls = min_calls

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if self._value == UNDEFINED or _significantly_changed(self._value, new_value):
            self._calls += 1
        else:
            self._calls = 0

        if self._value == UNDEFINED or self._calls >= self._min_calls:
            self._value = (
                copy(new_value) if isinstance(new_value, Parameter) else new_value
            )
            self._calls = 0
            return await self._callback(new_value)


def debounce(callback: Callback, min_calls: int) -> _Debounce:
    """Return a debounce filter.

    A callback function will only called once value is stabilized
    across multiple filter calls.

    :param callback: A callback function to be awaited on value change
    :type callback: Callable[[Any], Coroutine[Any, Any, Any]]
    :param min_calls: Value shouldn't change for this amount of
        filter calls
    :type min_calls: int
    :return: A instance of callable filter
    :rtype: _Debounce
    """
    return _Debounce(callback, min_calls)


class _Throttle(Filter):
    """Represents a throttle filter.

    Calls a callback only when certain amount of seconds passed
    since the last call.
    """

    __slots__ = ("_last_called", "_timeout")

    _last_called: float | None
    _timeout: float

    def __init__(self, callback: Callback, seconds: float) -> None:
        """Initialize a new throttle filter."""
        super().__init__(callback)
        self._last_called = None
        self._timeout = seconds

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        current_timestamp = time.monotonic()
        if (
            self._last_called is None
            or (current_timestamp - self._last_called) >= self._timeout
        ):
            self._last_called = current_timestamp
            return await self._callback(new_value)


def throttle(callback: Callback, seconds: float) -> _Throttle:
    """Return a throttle filter.

    A callback function will only be called once a certain amount of
    seconds passed since the last call.

    :param callback: A callback function that will be awaited once
        filter conditions are fulfilled
    :type callback: Callable[[Any], Coroutine[Any, Any, Any]]
    :param seconds: A callback will be awaited at most once per
        this amount of seconds
    :type seconds: float
    :return: A instance of callable filter
    :rtype: _Throttle
    """
    return _Throttle(callback, seconds)


class _Delta(Filter):
    """Represents a difference filter.

    Calls a callback with a difference between two subsequent values.
    """

    __slots__ = ()

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if self._value == UNDEFINED or _significantly_changed(self._value, new_value):
            old_value = self._value
            self._value = (
                copy(new_value) if isinstance(new_value, Parameter) else new_value
            )
            if (
                self._value != UNDEFINED
                and (difference := _diffence_between(old_value, new_value)) is not None
            ):
                return await self._callback(difference)


def delta(callback: Callback) -> _Delta:
    """Return a difference filter.

    A callback function will be called with a difference between two
    subsequent value.

    :param callback: A callback function that will be awaited with
        difference between values in two subsequent calls
    :type callback: Callable[[Any], Coroutine[Any, Any, Any]]
    :return: A instance of callable filter
    :rtype: _Delta
    """
    return _Delta(callback)


class _Aggregate(Filter):
    """Represents an aggregate filter.

    Calls a callback with a sum of values collected over a specified
    time period.
    """

    __slots__ = ("_sum", "_last_update", "_timeout")

    _sum: complex
    _last_update: float
    _timeout: float

    def __init__(self, callback: Callback, seconds: float) -> None:
        """Initialize a new aggregate filter."""
        super().__init__(callback)
        self._last_update = time.monotonic()
        self._timeout = seconds
        self._sum = 0.0

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        current_timestamp = time.monotonic()
        try:
            self._sum += new_value
        except TypeError as e:
            raise ValueError(
                "Aggregate filter can only be used with numeric values"
            ) from e

        if current_timestamp - self._last_update >= self._timeout:
            result = await self._callback(self._sum)
            self._last_update = current_timestamp
            self._sum = 0.0
            return result


def aggregate(callback: Callback, seconds: float) -> _Aggregate:
    """Return an aggregate filter.

    A callback function will be called with a sum of values collected
    over a specified time period. Can only be used with numeric values.

    :param callback: A callback function to be awaited once filter
        conditions are fulfilled
    :type callback: Callable[[Any], Coroutine[Any, Any, Any]]
    :param seconds: A callback will be awaited with a sum of values
        aggregated over this amount of seconds.
    :type seconds: float
    :return: A instance of callable filter
    :rtype: _Aggregate
    """
    return _Aggregate(callback, seconds)


class _Custom(Filter):
    """Represents a custom filter.

    Calls a callback with value, if user-defined filter function
    that's called by this class with the value as an argument
    returns true.
    """

    __slots__ = ("_filter_fn",)

    filter_fn: Callable[[Any], bool]

    def __init__(self, callback: Callback, filter_fn: Callable[[Any], bool]) -> None:
        """Initialize a new custom filter."""
        super().__init__(callback)
        self._filter_fn = filter_fn

    async def __call__(self, new_value: Any) -> Any:
        """Set a new value for the callback."""
        if self._filter_fn(new_value):
            await self._callback(new_value)


def custom(callback: Callback, filter_fn: Callable[[Any], bool]) -> _Custom:
    """Return a custom filter.

    A callback function will be called when user-defined filter
    function, that's being called with the value as an argument,
    returns true.

    :param callback: A callback function to be awaited when
        filter function return true
    :type callback: Callable[[Any], Coroutine[Any, Any, Any]]
    :param filter_fn: Filter function, that will be called with a
        value and should return `True` to await filter's callback
    :type filter_fn: Callable[[Any], bool]
    :return: A instance of callable filter
    :rtype: _Custom
    """
    return _Custom(callback, filter_fn)



"""Contains request frames."""

from __future__ import annotations

from typing import Any, ClassVar

from pyplumio.const import (
    ATTR_COUNT,
    ATTR_DEVICE_INDEX,
    ATTR_INDEX,
    ATTR_OFFSET,
    ATTR_SIZE,
    ATTR_START,
    ATTR_VALUE,
    FrameType,
)
from pyplumio.exceptions import FrameDataError
from pyplumio.frames import Request, Response
from pyplumio.frames.responses import DeviceAvailableResponse, ProgramVersionResponse
from pyplumio.structures.schedules import SchedulesStructure


class ProgramVersionRequest(Request):
    """Represents a program version request."""

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_PROGRAM_VERSION

    def response(self, **kwargs: Any) -> Response | None:
        """Return a response frame."""
        return ProgramVersionResponse(recipient=self.sender, **kwargs)


class CheckDeviceRequest(Request):
    """Represents a check device request."""

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_CHECK_DEVICE

    def response(self, **kwargs: Any) -> Response | None:
        """Return a response frame."""
        return DeviceAvailableResponse(recipient=self.sender, **kwargs)


class UIDRequest(Request):
    """Represents an UID request."""

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_UID


class PasswordRequest(Request):
    """Represents a password request."""

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_PASSWORD


class EcomaxParametersRequest(Request):
    """Represents an ecoMAX parameters request.

    Contains number of parameters and index of the first parameter.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_ECOMAX_PARAMETERS

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        return bytearray([data.get(ATTR_COUNT, 255), data.get(ATTR_START, 0)])


class MixerParametersRequest(Request):
    """Represents a mixer parameters request.

    Contains number of parameters to get and index of the first
    parameter.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_MIXER_PARAMETERS

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        return bytearray([data.get(ATTR_COUNT, 255), data.get(ATTR_START, 0)])


class ThermostatParametersRequest(Request):
    """Represents a thermostat parameters request.

    Contains number of parameters to get and index of the first
    parameter.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_THERMOSTAT_PARAMETERS

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        return bytearray([data.get(ATTR_COUNT, 255), data.get(ATTR_START, 0)])


class RegulatorDataSchemaRequest(Request):
    """Represents regulator data schema request."""

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_REGULATOR_DATA_SCHEMA


class SetEcomaxParameterRequest(Request):
    """Represents a request to set an ecoMAX parameter.

    Contains parameter index and new value.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_SET_ECOMAX_PARAMETER

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        try:
            return bytearray([data[ATTR_INDEX], data[ATTR_VALUE]])
        except (KeyError, ValueError) as e:
            raise FrameDataError from e


class SetMixerParameterRequest(Request):
    """Represents a request to set a mixer parameter.

    Contains parameter index, new value and mixer index.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_SET_MIXER_PARAMETER

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        try:
            return bytearray(
                [data[ATTR_DEVICE_INDEX], data[ATTR_INDEX], data[ATTR_VALUE]]
            )
        except (KeyError, ValueError) as e:
            raise FrameDataError from e


class SetThermostatParameterRequest(Request):
    """Represents a request to set a thermostat parameter.

    Contains parameter index, new value, parameter offset and
    parameter size.

    Parameter offset is calculated by multiplying
    thermostat index by total number of parameters available.

    For example, if total available parameters is 12,
    offset for the second thermostat is 12 (1 * 12),
    and for the third thermostat is 24 (2 * 12).
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_SET_THERMOSTAT_PARAMETER

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        try:
            index = data[ATTR_INDEX]
            value: int = data[ATTR_VALUE]
            offset = data[ATTR_OFFSET]
            message = bytearray([index if offset is None else index + offset])
            return message + value.to_bytes(length=data[ATTR_SIZE], byteorder="little")
        except (KeyError, ValueError) as e:
            raise FrameDataError from e


class EcomaxControlRequest(Request):
    """Represents an ecoMAX control request.

    Contains single binary value. 0 - means that controller should
    be turned off, 1 - means that it should be turned on.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_ECOMAX_CONTROL

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        try:
            return bytearray([data[ATTR_VALUE]])
        except KeyError as e:
            raise FrameDataError from e


class StartMasterRequest(Request):
    """Represents a request to become a master.

    Once controller receives this request, it starts sending
    periodic messages.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_START_MASTER


class StopMasterRequest(Request):
    """Represents a request to stop being a master.

    Once controller receives this request, it stops sending
    periodic messages.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_STOP_MASTER


class AlertsRequest(Request):
    """Represents an alerts request.

    Contains number of alerts to get and index of the first
    alert.
    """

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_ALERTS

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        return bytearray([data.get(ATTR_START, 0), data.get(ATTR_COUNT, 10)])


class SchedulesRequest(Request):
    """Represents a schedules request."""

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_SCHEDULES


class SetScheduleRequest(Request):
    """Represents a request to set a schedule."""

    __slots__ = ()

    frame_type: ClassVar[FrameType] = FrameType.REQUEST_SET_SCHEDULE

    def create_message(self, data: dict[str, Any]) -> bytearray:
        """Create a frame message."""
        return SchedulesStructure(self).encode(data)



"""Contains a fuel level structure decoder."""

from __future__ import annotations

from typing import Any, Final

from pyplumio.const import BYTE_UNDEFINED
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_FUEL_LEVEL: Final = "fuel_level"

FUEL_LEVEL_OFFSET: Final = 101


class FuelLevelStructure(StructureDecoder):
    """Represents a fuel level sensor data structure."""

    __slots__ = ()

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        fuel_level = message[offset]
        offset += 1

        if fuel_level == BYTE_UNDEFINED:
            return ensure_dict(data), offset

        if fuel_level >= FUEL_LEVEL_OFFSET:
            # Observed on at least ecoMAX 860P6-O.
            # See: https://github.com/denpamusic/PyPlumIO/issues/19
            fuel_level -= FUEL_LEVEL_OFFSET

        return (ensure_dict(data, {ATTR_FUEL_LEVEL: fuel_level}), offset)



"""Contains tests for the UID helper class."""

import pytest

from pyplumio.helpers.uid import decode_uid


@pytest.mark.parametrize(
    ("message", "uid"),
    [
        ("001600110D383338365539", "D251PAKR3GCPZ1K8G05G0"),
        ("002500300E191932135831", "CE71HB09J468P1ZZ00980"),
    ],
)
def test_from_bytes(message, uid) -> None:
    """Test unpacking an UID from bytes."""
    assert decode_uid(bytes.fromhex(message)) == uid



"""Contains a statuses structure decoder."""

from __future__ import annotations

from typing import Any, Final

from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_HEATING_TARGET: Final = "heating_target"
ATTR_HEATING_STATUS: Final = "heating_status"
ATTR_WATER_HEATER_TARGET: Final = "water_heater_target"
ATTR_WATER_HEATER_STATUS: Final = "water_heater_status"

STATUSES: tuple[str, ...] = (
    ATTR_HEATING_TARGET,
    ATTR_HEATING_STATUS,
    ATTR_WATER_HEATER_TARGET,
    ATTR_WATER_HEATER_STATUS,
)

STATUSES_SIZE: Final = 4


class StatusesStructure(StructureDecoder):
    """Represents a statuses data structure."""

    __slots__ = ()

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        return (
            ensure_dict(
                data,
                {
                    status: message[offset + index]
                    for index, status in enumerate(STATUSES)
                },
            ),
            offset + STATUSES_SIZE,
        )
